"""This module implements equivalents of the basic ComputedEntry objects, which
is the basic entity that can be used to perform many analyses. ComputedEntries
contain calculated information, typically from VASP or other electronic
structure codes. For example, ComputedEntries can be used as inputs for phase
diagram analysis.
"""

from __future__ import annotations

import abc
import json
import math
import os
import warnings
from itertools import combinations
from typing import TYPE_CHECKING, Literal, cast

import numpy as np
from monty.json import MontyDecoder, MontyEncoder, MSONable
from scipy.interpolate import interp1d
from uncertainties import ufloat

from pymatgen.core.composition import Composition
from pymatgen.entries import Entry
from pymatgen.util.due import Doi, due

if TYPE_CHECKING:
    from pymatgen.core import Structure

__author__ = "Ryan Kingsbury, Matt McDermott, Shyue Ping Ong, Anubhav Jain"
__copyright__ = "Copyright 2011-2020, The Materials Project"
__version__ = "1.1"
__date__ = "April 2020"

with open(os.path.join(os.path.dirname(__file__), "data/g_els.json")) as file:
    G_ELEMS = json.load(file)
with open(os.path.join(os.path.dirname(__file__), "data/nist_gas_gf.json")) as file:
    G_GASES = json.load(file)


class EnergyAdjustment(MSONable):
    """Lightweight class to contain information about an energy adjustment or
    energy correction.
    """

    def __init__(
        self,
        value,
        uncertainty=np.nan,
        name="Manual adjustment",
        cls=None,
        description="",
    ):
        """
        Args:
            value (float): value of the energy adjustment in eV
            uncertainty (float): uncertainty of the energy adjustment in eV. Default: np.nan
            name (str): human-readable name of the energy adjustment.
                (Default: Manual adjustment)
            cls (dict): Serialized Compatibility class used to generate the energy adjustment. Defaults to {}.
            description (str): human-readable explanation of the energy adjustment.
        """
        self.name = name
        self.cls = cls or {}
        self.description = description
        self._value = value
        self._uncertainty = uncertainty

    @property
    def value(self):
        """Return the value of the energy correction in eV."""
        return self._value

    @property
    def uncertainty(self):
        """Return the uncertainty in the value of the energy adjustment in eV."""
        return self._uncertainty

    @abc.abstractmethod
    def normalize(self, factor):
        """Scale the value of the current energy adjustment by factor in-place.

        This method is utilized in ComputedEntry.normalize() to scale the energies to a formula unit basis
        (e.g. E_Fe6O9 = 3 x E_Fe2O3).
        """

    @property
    @abc.abstractmethod
    def explain(self):
        """Return an explanation of how the energy adjustment is calculated."""

    def __repr__(self):
        name, value, uncertainty, description = self.name, float(self.value), self.uncertainty, self.description
        # self.cls might not be a dict if monty decoding is enabled in the new MPRester
        # which hydrates all dicts with @class and @module keys into classes in which case
        # we expect a Compatibility subclass
        generated_by = self.cls.get("@class", "unknown") if isinstance(self.cls, dict) else type(self.cls).__name__
        return f"{type(self).__name__}({name=}, {value=:.3}, {uncertainty=:.3}, {description=}, {generated_by=})"


class ConstantEnergyAdjustment(EnergyAdjustment):
    """A constant energy adjustment applied to a ComputedEntry. Useful in energy referencing
    schemes such as the Aqueous energy referencing scheme.
    """

    def __init__(
        self,
        value,
        uncertainty=np.nan,
        name="Constant energy adjustment",
        cls=None,
        description="Constant energy adjustment",
    ):
        """
        Args:
            value: float, value of the energy adjustment in eV
            uncertainty: float, uncertainty of the energy adjustment in eV. (Default: np.nan)
            name: str, human-readable name of the energy adjustment.
                (Default: Constant energy adjustment)
            cls: dict, Serialized Compatibility class used to generate the energy
                adjustment. (Default: None)
            description: str, human-readable explanation of the energy adjustment.
        """
        super().__init__(value, uncertainty, name=name, cls=cls, description=description)
        self._value = value
        self._uncertainty = uncertainty

    @property
    def explain(self):
        """Return an explanation of how the energy adjustment is calculated."""
        return f"{self.description} ({self.value:.3f} eV)"

    def normalize(self, factor):
        """Normalize energy adjustment (in place), dividing value/uncertainty by a
        factor.
        :param factor: factor to divide by.
        """
        self._value /= factor
        self._uncertainty /= factor


class ManualEnergyAdjustment(ConstantEnergyAdjustment):
    """A manual energy adjustment applied to a ComputedEntry."""

    def __init__(self, value):
        """
        Args:
            value: float, value of the energy adjustment in eV.
        """
        name = "Manual energy adjustment"
        description = "Manual energy adjustment"
        super().__init__(value, name=name, cls=None, description=description)


class CompositionEnergyAdjustment(EnergyAdjustment):
    """An energy adjustment applied to a ComputedEntry based on the atomic composition.
    Used in various DFT energy correction schemes.
    """

    def __init__(
        self,
        adj_per_atom,
        n_atoms,
        uncertainty_per_atom=np.nan,
        name="",
        cls=None,
        description="Composition-based energy adjustment",
    ):
        """
        Args:
            adj_per_atom: float, energy adjustment to apply per atom, in eV/atom
            n_atoms: float or int, number of atoms.
            uncertainty_per_atom: float, uncertainty in energy adjustment to apply per atom, in eV/atom.
                (Default: np.nan)
            name: str, human-readable name of the energy adjustment.
                (Default: "")
            cls: dict, Serialized Compatibility class used to generate the energy
                adjustment. (Default: None)
            description: str, human-readable explanation of the energy adjustment.
        """
        self._adj_per_atom = adj_per_atom
        self.uncertainty_per_atom = uncertainty_per_atom
        self.n_atoms = n_atoms
        self.cls = cls or {}
        self.name = name
        self.description = description

    @property
    def value(self):
        """Return the value of the energy adjustment in eV."""
        return self._adj_per_atom * self.n_atoms

    @property
    def uncertainty(self):
        """Return the value of the energy adjustment in eV."""
        return self.uncertainty_per_atom * self.n_atoms

    @property
    def explain(self):
        """Return an explanation of how the energy adjustment is calculated."""
        return f"{self.description} ({self._adj_per_atom:.3f} eV/atom x {self.n_atoms} atoms)"

    def normalize(self, factor):
        """Normalize energy adjustment (in place), dividing value/uncertainty by a
        factor.
        :param factor: factor to divide by.
        """
        self.n_atoms /= factor


class TemperatureEnergyAdjustment(EnergyAdjustment):
    """An energy adjustment applied to a ComputedEntry based on the temperature.
    Used, for example, to add entropy to DFT energies.
    """

    def __init__(
        self,
        adj_per_deg,
        temp,
        n_atoms,
        uncertainty_per_deg=np.nan,
        name="",
        cls=None,
        description="Temperature-based energy adjustment",
    ):
        """
        Args:
            adj_per_deg: float, energy adjustment to apply per degree K, in eV/atom
            temp: float, temperature in Kelvin
            n_atoms: float or int, number of atoms
            uncertainty_per_deg: float, uncertainty in energy adjustment to apply per degree K,
                in eV/atom. (Default: np.nan)
            name: str, human-readable name of the energy adjustment.
                (Default: "")
            cls: dict, Serialized Compatibility class used to generate the energy
                adjustment. (Default: None)
            description: str, human-readable explanation of the energy adjustment.
        """
        self._adj_per_deg = adj_per_deg
        self.uncertainty_per_deg = uncertainty_per_deg
        self.temp = temp
        self.n_atoms = n_atoms
        self.name = name
        self.cls = cls or {}
        self.description = description

    @property
    def value(self):
        """Return the value of the energy correction in eV."""
        return self._adj_per_deg * self.temp * self.n_atoms

    @property
    def uncertainty(self):
        """Return the value of the energy adjustment in eV."""
        return self.uncertainty_per_deg * self.temp * self.n_atoms

    @property
    def explain(self):
        """Return an explanation of how the energy adjustment is calculated."""
        return f"{self.description} ({self._adj_per_deg:.4f} eV/K/atom x {self.temp} K x {self.n_atoms} atoms)"

    def normalize(self, factor):
        """Normalize energy adjustment (in place), dividing value/uncertainty by a
        factor.
        :param factor: factor to divide by.
        """
        self.n_atoms /= factor


class ComputedEntry(Entry):
    """Lightweight Entry object for computed data. Contains facilities
    for applying corrections to the energy attribute and for storing
    calculation parameters.
    """

    def __init__(
        self,
        composition: Composition | str | dict[str, float],
        energy: float,
        correction: float = 0.0,
        energy_adjustments: list | None = None,
        parameters: dict | None = None,
        data: dict | None = None,
        entry_id: object | None = None,
    ):
        """Initializes a ComputedEntry.

        Args:
            composition (Composition): Composition of the entry. For
                flexibility, this can take the form of all the typical input
                taken by a Composition, including a {symbol: amt} dict,
                a string formula, and others.
            energy (float): Energy of the entry. Usually the final calculated
                energy from VASP or other electronic structure codes.
            correction (float): Manually set an energy correction, will ignore
                energy_adjustments if specified.
            energy_adjustments: An optional list of EnergyAdjustment to
                be applied to the energy. This is used to modify the energy for
                certain analyses. Defaults to None.
            parameters: An optional dict of parameters associated with
                the entry. Defaults to None.
            data: An optional dict of any additional data associated
                with the entry. Defaults to None.
            entry_id: An optional id to uniquely identify the entry.
        """
        super().__init__(composition, energy)
        self.energy_adjustments = energy_adjustments or []

        if correction != 0.0:
            if energy_adjustments:
                raise ValueError(
                    f"Argument conflict! Setting correction = {correction:.3f} conflicts "
                    "with setting energy_adjustments. Specify one or the "
                    "other."
                )

            self.correction = correction

        self.parameters = parameters or {}
        self.data = data or {}
        self.entry_id = entry_id
        self.name = self.composition.reduced_formula

    @property
    def uncorrected_energy(self) -> float:
        """
        Returns:
            float: the *uncorrected* energy of the entry.
        """
        return self._energy

    @property
    def energy(self) -> float:
        """The *corrected* energy of the entry."""
        return self.uncorrected_energy + self.correction

    @property
    def uncorrected_energy_per_atom(self) -> float:
        """
        Returns:
            float: the *uncorrected* energy of the entry, normalized by atoms in eV/atom.
        """
        return self.uncorrected_energy / self.composition.num_atoms

    @property
    def correction(self) -> float:
        """
        Returns:
            float: the total energy correction / adjustment applied to the entry in eV.
        """
        # adds to ufloat(0.0, 0.0) to ensure that no corrections still result in ufloat object
        corr = ufloat(0.0, 0.0) + sum(ufloat(ea.value, ea.uncertainty) for ea in self.energy_adjustments)
        return corr.nominal_value

    @correction.setter
    def correction(self, x: float) -> None:
        corr = ManualEnergyAdjustment(x)
        self.energy_adjustments = [corr]

    @property
    def correction_per_atom(self) -> float:
        """
        Returns:
            float: the total energy correction / adjustment applied to the entry in eV/atom.
        """
        return self.correction / self.composition.num_atoms

    @property
    def correction_uncertainty(self) -> float:
        """
        Returns:
            float: the uncertainty of the energy adjustments applied to the entry in eV.
        """
        # adds to ufloat(0.0, 0.0) to ensure that no corrections still result in ufloat object
        unc = ufloat(0.0, 0.0) + sum(
            ufloat(ea.value, ea.uncertainty) if not np.isnan(ea.uncertainty) else ufloat(ea.value, 0)
            for ea in self.energy_adjustments
        )

        if unc.nominal_value != 0 and unc.std_dev == 0:
            return np.nan

        return unc.std_dev

    @property
    def correction_uncertainty_per_atom(self) -> float:
        """
        Returns:
            float: the uncertainty of the energy adjustments applied to the entry in eV/atom.
        """
        return self.correction_uncertainty / self.composition.num_atoms

    def normalize(self, mode: Literal["formula_unit", "atom"] = "formula_unit") -> ComputedEntry:
        """Normalize the entry's composition and energy.

        Args:
            mode ("formula_unit" | "atom"): "formula_unit" (the default) normalizes to composition.reduced_formula.
                "atom" normalizes such that the composition amounts sum to 1.
        """
        factor = self._normalization_factor(mode)
        new_composition = self._composition / factor
        new_energy = self._energy / factor

        new_entry_dict = self.as_dict()
        new_entry_dict["composition"] = new_composition.as_dict()
        new_entry_dict["energy"] = new_energy

        # TODO: make sure EnergyAdjustments are _also_ immutable to avoid this hacking
        new_energy_adjustments = MontyDecoder().process_decoded(new_entry_dict["energy_adjustments"])
        for ea in new_energy_adjustments:
            ea.normalize(factor)
        new_entry_dict["energy_adjustments"] = [ea.as_dict() for ea in new_energy_adjustments]

        return self.from_dict(new_entry_dict)

    def __repr__(self) -> str:
        n_atoms = self.composition.num_atoms
        output = [
            f"{self.entry_id} {type(self).__name__:<10} "
            f"- {self.composition.formula:<12} ({self.composition.reduced_formula})",
            f"{'Energy (Uncorrected)':<24} = {self._energy:<9.4f} eV ({self._energy / n_atoms:<8.4f} eV/atom)",
            f"{'Correction':<24} = {self.correction:<9.4f} eV ({self.correction / n_atoms:<8.4f} eV/atom)",
            f"{'Energy (Final)':<24} = {self.energy:<9.4f} eV ({self.energy_per_atom:<8.4f} eV/atom)",
            "Energy Adjustments:",
        ]
        if len(self.energy_adjustments) == 0:
            output.append("  None")
        else:
            for e in self.energy_adjustments:
                output.append(f"  {e.name:<23}: {e.value:<9.4f} eV ({e.value / n_atoms:<8.4f} eV/atom)")
        output.append("Parameters:")
        for k, v in self.parameters.items():
            output.append(f"  {k:<22} = {v}")
        output.append("Data:")
        for k, v in self.data.items():
            output.append(f"  {k:<22} = {v}")
        return "\n".join(output)

    def __eq__(self, other: object) -> bool:
        # NOTE: Scaled duplicates i.e. physically equivalent materials
        # are not equal unless normalized separately.
        if self is other:
            return True

        needed_attrs = ("composition", "energy", "entry_id")
        if not all(hasattr(other, attr) for attr in needed_attrs):
            return NotImplemented

        other = cast(ComputedEntry, other)

        # Equality is defined based on composition and energy.
        # If structures are involved, it is assumed that a {composition, energy} is
        # vanishingly unlikely to be the same if the structures are different.
        # If entry_ids are different, assume that the entries are different.
        # However, if entry_id is same, they may have different corrections (e.g., due
        # to mixing scheme used) and thus should be compared on corrected energy.

        if getattr(self, "entry_id", None) and getattr(other, "entry_id", None) and self.entry_id != other.entry_id:
            return False

        if not math.isclose(self.energy, other.energy):
            return False

        if self.composition != other.composition:
            return False

        # assumes that data, parameters are equivalent
        return True

    @classmethod
    def from_dict(cls, d) -> ComputedEntry:
        """:param d: Dict representation.

        Returns:
            ComputedEntry
        """
        dec = MontyDecoder()
        # the first block here is for legacy ComputedEntry that were
        # serialized before we had the energy_adjustments attribute.
        if d["correction"] != 0 and not d.get("energy_adjustments"):
            return cls(
                d["composition"],
                d["energy"],
                d["correction"],
                parameters={k: dec.process_decoded(v) for k, v in d.get("parameters", {}).items()},
                data={k: dec.process_decoded(v) for k, v in d.get("data", {}).items()},
                entry_id=d.get("entry_id"),
            )
        # this is the preferred / modern way of instantiating ComputedEntry
        # we don't pass correction explicitly because it will be calculated
        # on the fly from energy_adjustments
        return cls(
            d["composition"],
            d["energy"],
            correction=0,
            energy_adjustments=[dec.process_decoded(e) for e in d.get("energy_adjustments", {})],
            parameters={k: dec.process_decoded(v) for k, v in d.get("parameters", {}).items()},
            data={k: dec.process_decoded(v) for k, v in d.get("data", {}).items()},
            entry_id=d.get("entry_id"),
        )

    def as_dict(self) -> dict:
        """MSONable dict."""
        return_dict = super().as_dict()
        return_dict.update(
            {
                "entry_id": self.entry_id,
                "correction": self.correction,
                "energy_adjustments": json.loads(json.dumps(self.energy_adjustments, cls=MontyEncoder)),
                "parameters": json.loads(json.dumps(self.parameters, cls=MontyEncoder)),
                "data": json.loads(json.dumps(self.data, cls=MontyEncoder)),
            }
        )
        return return_dict

    def __hash__(self) -> int:
        # NOTE It is assumed that the user will ensure entry_id is a
        # unique identifier for ComputedEntry type classes.
        if self.entry_id is not None:
            return hash(f"{type(self).__name__}{self.entry_id}")

        return super().__hash__()

    def copy(self) -> ComputedEntry:
        """Returns a copy of the ComputedEntry."""
        return ComputedEntry(
            composition=self.composition,
            energy=self.uncorrected_energy,
            energy_adjustments=self.energy_adjustments,
            parameters=self.parameters,
            data=self.data,
            entry_id=self.entry_id,
        )


class ComputedStructureEntry(ComputedEntry):
    """A heavier version of ComputedEntry which contains a structure as well. The
    structure is needed for some analyses.
    """

    def __init__(
        self,
        structure: Structure,
        energy: float,
        correction: float = 0.0,
        composition: Composition | str | dict[str, float] | None = None,
        energy_adjustments: list | None = None,
        parameters: dict | None = None,
        data: dict | None = None,
        entry_id: object | None = None,
    ) -> None:
        """Initializes a ComputedStructureEntry.

        Args:
            structure (Structure): The actual structure of an entry.
            energy (float): Energy of the entry. Usually the final calculated
                energy from VASP or other electronic structure codes.
            correction (float, optional): A correction to the energy. This is mutually exclusive with
                energy_adjustments, i.e. pass either or neither but not both. Defaults to 0.
            composition (Composition): Composition of the entry. For
                flexibility, this can take the form of all the typical input
                taken by a Composition, including a {symbol: amt} dict,
                a string formula, and others.
            energy_adjustments: An optional list of EnergyAdjustment to
                be applied to the energy. This is used to modify the energy for
                certain analyses. Defaults to None.
            parameters: An optional dict of parameters associated with
                the entry. Defaults to None.
            data: An optional dict of any additional data associated
                with the entry. Defaults to None.
            entry_id: An optional id to uniquely identify the entry.
        """
        if composition:
            composition = Composition(composition)
            if (
                composition.get_integer_formula_and_factor()[0]
                != structure.composition.get_integer_formula_and_factor()[0]
            ):
                raise ValueError("Mismatching composition provided.")
        else:
            composition = structure.composition

        super().__init__(
            composition,
            energy,
            correction=correction,
            energy_adjustments=energy_adjustments,
            parameters=parameters,
            data=data,
            entry_id=entry_id,
        )
        self._structure = structure

    @property
    def structure(self) -> Structure:
        """The structure of the entry."""
        return self._structure

    def as_dict(self) -> dict:
        """MSONable dict."""
        d = super().as_dict()
        d["structure"] = self.structure.as_dict()
        return d

    @classmethod
    def from_dict(cls, d) -> ComputedStructureEntry:
        """:param d: Dict representation.

        Returns:
            ComputedStructureEntry
        """
        dec = MontyDecoder()
        # the first block here is for legacy ComputedEntry that were
        # serialized before we had the energy_adjustments attribute.
        if d["correction"] != 0 and not d.get("energy_adjustments"):
            struct = dec.process_decoded(d["structure"])
            return cls(
                struct,
                d["energy"],
                correction=d["correction"],
                parameters={k: dec.process_decoded(v) for k, v in d.get("parameters", {}).items()},
                data={k: dec.process_decoded(v) for k, v in d.get("data", {}).items()},
                entry_id=d.get("entry_id"),
            )
        # this is the preferred / modern way of instantiating ComputedEntry
        # we don't pass correction explicitly because it will be calculated
        # on the fly from energy_adjustments
        return cls(
            dec.process_decoded(d["structure"]),
            d["energy"],
            composition=d.get("composition"),
            correction=0,
            energy_adjustments=[dec.process_decoded(e) for e in d.get("energy_adjustments", {})],
            parameters={k: dec.process_decoded(v) for k, v in d.get("parameters", {}).items()},
            data={k: dec.process_decoded(v) for k, v in d.get("data", {}).items()},
            entry_id=d.get("entry_id"),
        )

    def normalize(self, mode: Literal["formula_unit", "atom"] = "formula_unit") -> ComputedStructureEntry:
        """Normalize the entry's composition and energy. The structure remains unchanged.

        Args:
            mode ("formula_unit" | "atom"): "formula_unit" (the default) normalizes to composition.reduced_formula.
                "atom" normalizes such that the composition amounts sum to 1.
        """
        # TODO: this should raise TypeError since normalization does not make sense
        # raise TypeError("You cannot normalize a structure.")
        warnings.warn(
            f"Normalization of a `{type(self).__name__}` makes "
            "`self.composition` and `self.structure.composition` inconsistent"
            " - please use self.composition for all further calculations."
        )
        # TODO: find a better solution for creating copies instead of as/from dict
        factor = self._normalization_factor(mode)
        d = super().normalize(mode).as_dict()
        d["structure"] = self.structure.as_dict()
        entry = self.from_dict(d)
        entry._composition /= factor  # pylint: disable=E1101
        return entry

    def copy(self) -> ComputedStructureEntry:
        """Returns a copy of the ComputedStructureEntry."""
        return ComputedStructureEntry(
            structure=self.structure.copy(),
            energy=self.uncorrected_energy,
            composition=self.composition,
            energy_adjustments=self.energy_adjustments,
            parameters=self.parameters,
            data=self.data,
            entry_id=self.entry_id,
        )


class GibbsComputedStructureEntry(ComputedStructureEntry):
    """An extension to ComputedStructureEntry which includes the estimated Gibbs
    free energy of formation via a machine-learned model.
    """

    def __init__(
        self,
        structure: Structure,
        formation_enthalpy_per_atom: float,
        temp: float = 300,
        gibbs_model: Literal["SISSO"] = "SISSO",
        composition: Composition | None = None,
        correction: float = 0.0,
        energy_adjustments: list | None = None,
        parameters: dict | None = None,
        data: dict | None = None,
        entry_id: object | None = None,
    ):
        """
        Args:
            structure (Structure): The pymatgen Structure object of an entry.
            formation_enthalpy_per_atom (float): Formation enthalpy of the entry;
            must be
                calculated using phase diagram construction (eV)
            temp (float): Temperature in Kelvin. If temperature is not selected from
                one of [300, 400, 500, ... 2000 K], then free energies will
                be interpolated. Defaults to 300 K.
            gibbs_model ('SISSO'): Model for Gibbs Free energy. "SISSO", the descriptor
                created by Bartel et al. (2018) -- see reference in documentation, is
                currently the only supported option.
            composition (Composition): The composition of the entry. Defaults to None.
            correction (float): A correction to be applied to the energy. Defaults to 0.
            energy_adjustments (list): A list of energy adjustments to be applied to
                the energy. Defaults to None.
            parameters (dict): An optional dict of parameters associated with
                the entry. Defaults to None.
            data (dict): An optional dict of any additional data associated
                with the entry. Defaults to None.
            entry_id: An optional id to uniquely identify the entry.
        """
        if temp < 300 or temp > 2000:
            raise ValueError("Temperature must be selected from range: [300, 2000] K.")

        integer_formula, _ = structure.composition.get_integer_formula_and_factor()

        self.experimental = False
        if integer_formula in G_GASES:
            self.experimental = True
            if "Experimental" not in str(entry_id):
                entry_id = f"{entry_id} (Experimental)"

        super().__init__(
            structure,
            energy=0,  # placeholder, energy reassigned at end of __init__
            composition=composition,
            correction=correction,
            energy_adjustments=energy_adjustments,
            parameters=parameters,
            data=data,
            entry_id=entry_id,
        )

        self.temp = temp
        self.gibbs_model = gibbs_model
        self.formation_enthalpy_per_atom = formation_enthalpy_per_atom

        self.interpolated = False
        if self.temp % 100:
            self.interpolated = True

        if gibbs_model.lower() == "sisso":
            self.gibbs_fn = self.gf_sisso
        else:
            raise ValueError(f"{gibbs_model} not a valid model. The only currently available model is 'SISSO'.")

        self._energy = self.gibbs_fn()

    @due.dcite(Doi("10.1038/s41467-018-06682-4", "Gibbs free energy SISSO descriptor"))
    def gf_sisso(self) -> float:
        """Gibbs Free Energy of formation as calculated by SISSO descriptor from Bartel
        et al. (2018). Units: eV (not normalized).

        WARNING: This descriptor only applies to solids. The implementation here
        attempts to detect and use downloaded NIST-JANAF data for common
        experimental gases (e.g. CO2) where possible. Note that experimental data is
        only for Gibbs Free Energy of formation, so expt. entries will register as
        having a formation enthalpy of 0.

        Reference: Bartel, C. J., Millican, S. L., Deml, A. M., Rumptz, J. R.,
        Tumas, W., Weimer, A. W., … Holder, A. M. (2018). Physical descriptor for
        the Gibbs energy of inorganic crystalline solids and
        temperature-dependent materials chemistry. Nature Communications, 9(1),
        4168. https://doi.org/10.1038/s41467-018-06682-4

        Returns:
            float: the difference between formation enthalpy (T=0 K, Materials
            Project) and the predicted Gibbs free energy of formation  (eV)
        """
        comp = self.composition

        if comp.is_element:
            return 0

        integer_formula, factor = comp.get_integer_formula_and_factor()
        if self.experimental:
            data = G_GASES[integer_formula]

            if self.interpolated:
                g_interp = interp1d([int(t) for t in data], list(data.values()))
                energy = g_interp(self.temp)
            else:
                energy = data[str(self.temp)]

            gibbs_energy = energy * factor
        else:
            num_atoms = len(self.structure)
            vol_per_atom = self.structure.volume / num_atoms
            reduced_mass = self._reduced_mass(self.structure)

            gibbs_energy = (
                comp.num_atoms
                * (self.formation_enthalpy_per_atom + self._g_delta_sisso(vol_per_atom, reduced_mass, self.temp))
                - self._sum_g_i()
            )

        return gibbs_energy

    def _sum_g_i(self) -> float:
        """Sum of the stoichiometrically weighted chemical potentials of the elements
        at specified temperature, as acquired from "g_els.json".

        Returns:
            float: sum of weighted chemical potentials [eV]
        """
        elems = self.composition.get_el_amt_dict()

        if self.interpolated:
            sum_g_i = 0
            for elem, amt in elems.items():
                g_interp = interp1d(
                    [float(t) for t in G_ELEMS],
                    [g_dict[elem] for g_dict in G_ELEMS.values()],
                )
                sum_g_i += amt * g_interp(self.temp)
        else:
            sum_g_i = sum(amt * G_ELEMS[str(self.temp)][elem] for elem, amt in elems.items())

        return sum_g_i

    @staticmethod
    def _reduced_mass(structure) -> float:
        """Reduced mass as calculated via Eq. 6 in Bartel et al. (2018).

        Args:
            structure (Structure): The pymatgen Structure object of the entry.

        Returns:
            float: reduced mass (amu)
        """
        reduced_comp = structure.composition.reduced_composition
        num_elems = len(reduced_comp.elements)
        elem_dict = reduced_comp.get_el_amt_dict()

        denominator = (num_elems - 1) * reduced_comp.num_atoms

        all_pairs = combinations(elem_dict.items(), 2)
        mass_sum = 0

        for pair in all_pairs:
            m_i = Composition(pair[0][0]).weight
            m_j = Composition(pair[1][0]).weight
            alpha_i = pair[0][1]
            alpha_j = pair[1][1]

            mass_sum += (alpha_i + alpha_j) * (m_i * m_j) / (m_i + m_j)

        return (1 / denominator) * mass_sum

    @staticmethod
    def _g_delta_sisso(vol_per_atom, reduced_mass, temp) -> float:
        """G^delta as predicted by SISSO-learned descriptor from Eq. (4) in
        Bartel et al. (2018).

        Args:
            vol_per_atom (float): volume per atom [Å^3/atom]
            reduced_mass (float): as calculated with pair-wise sum formula
                [amu]
            temp (float): Temperature [K]

        Returns:
            float: G^delta [eV/atom]
        """
        return (
            (-2.48e-4 * np.log(vol_per_atom) - 8.94e-5 * reduced_mass / vol_per_atom) * temp
            + 0.181 * np.log(temp)
            - 0.882
        )

    @classmethod
    def from_pd(cls, pd, temp=300, gibbs_model="SISSO") -> list[GibbsComputedStructureEntry]:
        """Constructor method for initializing a list of GibbsComputedStructureEntry
        objects from an existing T = 0 K phase diagram composed of
        ComputedStructureEntry objects, as acquired from a thermochemical database;
        (e.g.. The Materials Project).

        Args:
            pd (PhaseDiagram): T = 0 K phase diagram as created in pymatgen. Must
                contain ComputedStructureEntry objects.
            temp (int): Temperature [K] for estimating Gibbs free energy of formation.
            gibbs_model (str): Gibbs model to use; currently the only option is "SISSO".

        Returns:
            [GibbsComputedStructureEntry]: list of new entries which replace the orig.
                entries with inclusion of Gibbs free energy of formation at the
                specified temperature.
        """
        gibbs_entries = []
        for entry in pd.all_entries:
            if entry in pd.el_refs.values() or not entry.composition.is_element:
                gibbs_entries.append(
                    cls(
                        entry.structure,
                        formation_enthalpy_per_atom=pd.get_form_energy_per_atom(entry),
                        temp=temp,
                        correction=0,
                        gibbs_model=gibbs_model,
                        data=entry.data,
                        entry_id=entry.entry_id,
                    )
                )
        return gibbs_entries

    @classmethod
    def from_entries(cls, entries, temp=300, gibbs_model="SISSO") -> list[GibbsComputedStructureEntry]:
        """Constructor method for initializing GibbsComputedStructureEntry objects from
        T = 0 K ComputedStructureEntry objects, as acquired from a thermochemical
        database e.g. The Materials Project.

        Args:
            entries ([ComputedStructureEntry]): List of ComputedStructureEntry objects,
                as downloaded from The Materials Project API.
            temp (int): Temperature [K] for estimating Gibbs free energy of formation.
            gibbs_model (str): Gibbs model to use; currently the only option is "SISSO".

        Returns:
            [GibbsComputedStructureEntry]: list of new entries which replace the orig.
                entries with inclusion of Gibbs free energy of formation at the
                specified temperature.
        """
        from pymatgen.analysis.phase_diagram import PhaseDiagram

        pd = PhaseDiagram(entries)
        return cls.from_pd(pd, temp, gibbs_model)

    def as_dict(self) -> dict:
        """MSONable dict."""
        d = super().as_dict()
        d["formation_enthalpy_per_atom"] = self.formation_enthalpy_per_atom
        d["temp"] = self.temp
        d["gibbs_model"] = self.gibbs_model
        d["interpolated"] = self.interpolated
        return d

    @classmethod
    def from_dict(cls, d) -> GibbsComputedStructureEntry:
        """:param d: Dict representation.

        Returns:
            GibbsComputedStructureEntry
        """
        dec = MontyDecoder()
        return cls(
            dec.process_decoded(d["structure"]),
            d["formation_enthalpy_per_atom"],
            d["temp"],
            d["gibbs_model"],
            composition=d.get("composition"),
            correction=d["correction"],
            energy_adjustments=[dec.process_decoded(e) for e in d.get("energy_adjustments", {})],
            parameters={k: dec.process_decoded(v) for k, v in d.get("parameters", {}).items()},
            data={k: dec.process_decoded(v) for k, v in d.get("data", {}).items()},
            entry_id=d.get("entry_id"),
        )

    def __repr__(self):
        output = [
            f"GibbsComputedStructureEntry {self.entry_id} - {self.composition.formula}",
            f"Gibbs Free Energy (Formation) = {self.energy:.4f}",
        ]
        return "\n".join(output)



"""Classes for reading/manipulating/writing exciting input files."""

from __future__ import annotations

import xml.etree.ElementTree as ET

import numpy as np
import scipy.constants as const
from monty.io import zopen
from monty.json import MSONable

from pymatgen.core.lattice import Lattice
from pymatgen.core.periodic_table import Element
from pymatgen.core.structure import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.symmetry.bandstructure import HighSymmKpath

__author__ = "Christian Vorwerk"
__copyright__ = "Copyright 2016"
__version__ = "1.0"
__maintainer__ = "Christian Vorwerk"
__email__ = "vorwerk@physik.hu-berlin.de"
__status__ = "Development"
__date__ = "Nov 28, 2016"


class ExcitingInput(MSONable):
    """
    Object for representing the data stored in the structure part of the
    exciting input.

    .. attribute:: structure

        Associated Structure.

    .. attribute:: title

        Optional title string.

    .. attribute:: lockxyz

        Lockxyz attribute for each site if available. A Nx3 array of
        booleans.
    """

    def __init__(self, structure: Structure, title=None, lockxyz=None):
        """
        Args:
            structure (Structure):  Structure object.
            title (str): Optional title for exciting input. Defaults to unit
                cell formula of structure. Defaults to None.
            lockxyz (Nx3 array): bool values for selective dynamics,
                where N is number of sites. Defaults to None.
        """
        if structure.is_ordered:
            site_properties = {}
            if lockxyz:
                site_properties["selective_dynamics"] = lockxyz
            self.structure = structure.copy(site_properties=site_properties)
            self.title = structure.formula if title is None else title
        else:
            raise ValueError("Structure with partial occupancies cannot be converted into exciting input!")

    # define conversion factor between Bohr radius and Angstrom
    bohr2ang = const.value("Bohr radius") / const.value("Angstrom star")

    @property
    def lockxyz(self):
        """Selective dynamics site properties."""
        return self.structure.site_properties.get("selective_dynamics")

    @lockxyz.setter
    def lockxyz(self, lockxyz):
        self.structure.add_site_property("selective_dynamics", lockxyz)

    @classmethod
    @np.deprecate(message="Use from_str instead")
    def from_string(cls, *args, **kwargs):
        return cls.from_str(*args, **kwargs)

    @staticmethod
    def from_str(data):
        """Reads the exciting input from a string."""
        root = ET.fromstring(data)
        speciesnode = root.find("structure").iter("species")
        elements = []
        positions = []
        vectors = []
        lockxyz = []
        # get title
        title_in = str(root.find("title").text)
        # Read elements and coordinates
        for nodes in speciesnode:
            symbol = nodes.get("speciesfile").split(".")[0]
            if len(symbol.split("_")) == 2:
                symbol = symbol.split("_")[0]
            if Element.is_valid_symbol(symbol):
                # Try to recognize the element symbol
                element = symbol
            else:
                raise ValueError("Unknown element!")
            for atom in nodes.iter("atom"):
                x, y, z = atom.get("coord").split()
                positions.append([float(x), float(y), float(z)])
                elements.append(element)
                # Obtain lockxyz for each atom
                if atom.get("lockxyz") is not None:
                    lxyz = []
                    for line in atom.get("lockxyz").split():
                        if line in ("True", "true"):
                            lxyz.append(True)
                        else:
                            lxyz.append(False)
                    lockxyz.append(lxyz)
                else:
                    lockxyz.append([False, False, False])
        # check the atomic positions type
        if "cartesian" in root.find("structure").attrib:
            if root.find("structure").attrib["cartesian"]:
                cartesian = True
                for p in positions:
                    for j in range(3):
                        p[j] = p[j] * ExcitingInput.bohr2ang
                print(positions)
        else:
            cartesian = False
        # get the scale attribute
        scale_in = root.find("structure").find("crystal").get("scale")
        scale = float(scale_in) * ExcitingInput.bohr2ang if scale_in else ExcitingInput.bohr2ang
        # get the stretch attribute
        stretch_in = root.find("structure").find("crystal").get("stretch")
        stretch = np.array([float(a) for a in stretch_in]) if stretch_in else np.array([1.0, 1.0, 1.0])
        # get basis vectors and scale them accordingly
        basisnode = root.find("structure").find("crystal").iter("basevect")
        for vect in basisnode:
            x, y, z = vect.text.split()
            vectors.append(
                [
                    float(x) * stretch[0] * scale,
                    float(y) * stretch[1] * scale,
                    float(z) * stretch[2] * scale,
                ]
            )
        # create lattice and structure object
        lattice_in = Lattice(vectors)
        structure_in = Structure(lattice_in, elements, positions, coords_are_cartesian=cartesian)

        return ExcitingInput(structure_in, title_in, lockxyz)

    @staticmethod
    def from_file(filename):
        """
        :param filename: Filename

        Returns:
            ExcitingInput
        """
        with zopen(filename, "rt") as f:
            data = f.read().replace("\n", "")
        return ExcitingInput.from_str(data)

    def write_etree(self, celltype, cartesian=False, bandstr=False, symprec: float = 0.4, angle_tolerance=5, **kwargs):
        """
        Writes the exciting input parameters to an xml object.

        Args:
            celltype (str): Choice of unit cell. Can be either the unit cell
            from self.structure ("unchanged"), the conventional cell
            ("conventional"), or the primitive unit cell ("primitive").

            cartesian (bool): Whether the atomic positions are provided in
            Cartesian or unit-cell coordinates. Default is False.

            bandstr (bool): Whether the bandstructure path along the
            HighSymmKpath is included in the input file. Only supported if the
            celltype is set to "primitive". Default is False.

            symprec (float): Tolerance for the symmetry finding. Default is 0.4.

            angle_tolerance (float): Angle tolerance for the symmetry finding.
            Default is 5.

            **kwargs: Additional parameters for the input file.

        Returns:
            ET.Element containing the input XML structure
        """
        root = ET.Element("input")
        root.set(
            "{http://www.w3.org/2001/XMLSchema-instance}noNamespaceSchemaLocation",
            "http://xml.exciting-code.org/excitinginput.xsd",
        )
        title = ET.SubElement(root, "title")
        title.text = self.title
        if cartesian:
            structure = ET.SubElement(root, "structure", cartesian="true", speciespath="./")
        else:
            structure = ET.SubElement(root, "structure", speciespath="./")

        crystal = ET.SubElement(structure, "crystal")
        # set scale such that lattice vector can be given in Angstrom
        ang2bohr = const.value("Angstrom star") / const.value("Bohr radius")
        crystal.set("scale", str(ang2bohr))
        # determine which structure to use
        finder = SpacegroupAnalyzer(self.structure, symprec=symprec, angle_tolerance=angle_tolerance)
        if celltype == "primitive":
            new_struct = finder.get_primitive_standard_structure(international_monoclinic=False)
        elif celltype == "conventional":
            new_struct = finder.get_conventional_standard_structure(international_monoclinic=False)
        elif celltype == "unchanged":
            new_struct = self.structure
        else:
            raise ValueError("Type of unit cell not recognized!")

        # write lattice
        basis = new_struct.lattice.matrix
        for idx in range(3):
            basevect = ET.SubElement(crystal, "basevect")
            basevect.text = f"{basis[idx][0]:16.8f} {basis[idx][1]:16.8f} {basis[idx][2]:16.8f}"
        # write atomic positions for each species
        index = 0
        for elem in sorted(new_struct.types_of_species, key=lambda el: el.X):
            species = ET.SubElement(structure, "species", speciesfile=elem.symbol + ".xml")
            sites = new_struct.indices_from_symbol(elem.symbol)

            for j in sites:
                fc = new_struct[j].frac_coords
                coord = f"{fc[0]:16.8f} {fc[1]:16.8f} {fc[2]:16.8f}"
                # obtain Cartesian coords from fractional ones if needed
                if cartesian:
                    coord2 = []
                    for k in range(3):
                        inter = (
                            new_struct[j].frac_coords[k] * basis[0][k]
                            + new_struct[j].frac_coords[k] * basis[1][k]
                            + new_struct[j].frac_coords[k] * basis[2][k]
                        ) * ang2bohr
                        coord2.append(inter)
                    coord = f"{coord2[0]:16.8f} {coord2[1]:16.8f} {coord2[2]:16.8f}"

                # write atomic positions
                index = index + 1
                _ = ET.SubElement(species, "atom", coord=coord)
        # write bandstructure if needed
        if bandstr and celltype == "primitive":
            kpath = HighSymmKpath(new_struct, symprec=symprec, angle_tolerance=angle_tolerance)
            prop = ET.SubElement(root, "properties")
            bandstrct = ET.SubElement(prop, "bandstructure")
            for idx in range(len(kpath.kpath["path"])):
                plot = ET.SubElement(bandstrct, "plot1d")
                path = ET.SubElement(plot, "path", steps="100")
                for j in range(len(kpath.kpath["path"][idx])):
                    symbol = kpath.kpath["path"][idx][j]
                    coords = kpath.kpath["kpoints"][symbol]
                    coord = f"{coords[0]:16.8f} {coords[1]:16.8f} {coords[2]:16.8f}"
                    if symbol == "\\Gamma":
                        symbol = "GAMMA"
                    _ = ET.SubElement(path, "point", coord=coord, label=symbol)
        elif bandstr and celltype != "primitive":
            raise ValueError(
                "Bandstructure is only implemented for the \
                              standard primitive unit cell!"
            )

        # write extra parameters from kwargs if provided
        self._dicttoxml(kwargs, root)

        return root

    def write_string(self, celltype, cartesian=False, bandstr=False, symprec: float = 0.4, angle_tolerance=5, **kwargs):
        """
        Writes exciting input.xml as a string.

        Args:
            celltype (str): Choice of unit cell. Can be either the unit cell
            from self.structure ("unchanged"), the conventional cell
            ("conventional"), or the primitive unit cell ("primitive").

            cartesian (bool): Whether the atomic positions are provided in
            Cartesian or unit-cell coordinates. Default is False.

            bandstr (bool): Whether the bandstructure path along the
            HighSymmKpath is included in the input file. Only supported if the
            celltype is set to "primitive". Default is False.

            symprec (float): Tolerance for the symmetry finding. Default is 0.4.

            angle_tolerance (float): Angle tolerance for the symmetry finding.
            Default is 5.

            **kwargs: Additional parameters for the input file.

        Returns:
            String
        """
        try:
            root = self.write_etree(celltype, cartesian, bandstr, symprec, angle_tolerance, **kwargs)
            self._indent(root)
            # output should be a string not a bytes object
            string = ET.tostring(root).decode("UTF-8")
        except Exception:
            raise ValueError("Incorrect celltype!")
        return string

    def write_file(
        self, celltype, filename, cartesian=False, bandstr=False, symprec: float = 0.4, angle_tolerance=5, **kwargs
    ):
        """
        Writes exciting input file.

        Args:
            celltype (str): Choice of unit cell. Can be either the unit cell
            from self.structure ("unchanged"), the conventional cell
            ("conventional"), or the primitive unit cell ("primitive").

            filename (str): Filename for exciting input.

            cartesian (bool): Whether the atomic positions are provided in
            Cartesian or unit-cell coordinates. Default is False.

            bandstr (bool): Whether the bandstructure path along the
            HighSymmKpath is included in the input file. Only supported if the
            celltype is set to "primitive". Default is False.

            symprec (float): Tolerance for the symmetry finding. Default is 0.4.

            angle_tolerance (float): Angle tolerance for the symmetry finding.
            Default is 5.

            **kwargs: Additional parameters for the input file.
        """
        try:
            root = self.write_etree(celltype, cartesian, bandstr, symprec, angle_tolerance, **kwargs)
            self._indent(root)
            tree = ET.ElementTree(root)
            tree.write(filename)
        except Exception:
            raise ValueError("Incorrect celltype!")

    # Missing PrettyPrint option in the current version of xml.etree.cElementTree
    @staticmethod
    def _indent(elem, level=0):
        """
        Helper method to indent elements.

        :param elem:
        :param level:
        """
        i = "\n" + level * "  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for el in elem:
                ExcitingInput._indent(el, level + 1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        elif level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i

    def _dicttoxml(self, paramdict_, element):
        for key, value in paramdict_.items():
            if isinstance(value, str) and key == "text()":
                element.text = value
            elif isinstance(value, str):
                element.attrib[key] = value
            elif isinstance(value, list):
                for item in value:
                    self._dicttoxml(item, ET.SubElement(element, key))
            elif isinstance(value, dict):
                if element.findall(key) == []:
                    self._dicttoxml(value, ET.SubElement(element, key))
                else:
                    self._dicttoxml(value, element.findall(key)[0])
            else:
                print("cannot deal with", key, "=", value)



"""
This module defines the Cp2k output parser along with a few other functions for parsing cp2k-related
outputs.

"""

from __future__ import annotations

import logging
import os
import re
import warnings
from glob import glob
from itertools import chain

import numpy as np
import pandas as pd
from monty.io import zopen
from monty.json import MSONable, jsanitize
from monty.re import regrep

from pymatgen.core.structure import Molecule, Structure
from pymatgen.core.units import Ha_to_eV
from pymatgen.electronic_structure.bandstructure import BandStructure, BandStructureSymmLine
from pymatgen.electronic_structure.core import Orbital, Spin
from pymatgen.electronic_structure.dos import CompleteDos, Dos
from pymatgen.io.cp2k.inputs import Keyword
from pymatgen.io.cp2k.sets import Cp2kInput
from pymatgen.io.cp2k.utils import natural_keys, postprocessor
from pymatgen.io.xyz import XYZ

__author__ = "Nicholas Winner"
__version__ = "2.0"
__status__ = "Production"

logger = logging.getLogger(__name__)


class Cp2kOutput:
    """
    Class for parsing output file from CP2K. The CP2K output file is very flexible in the way that
    it is returned. This class will automatically parse parameters that should always be present,
    but other parsing features may be called depending on the run type.
    """

    def __init__(self, filename, verbose=False, auto_load=False):
        """
        Initialize the Cp2kOutput object.

        Args:
            filename: (str) Name of the CP2K output file to parse
            verbose: (bool) Whether or not to parse with verbosity (will parse lots of data that
                may not be useful)
            auto_load (bool): Whether or not to automatically load basic info like energies
                and structures.
        """
        # IO Info
        self.filename = filename
        self.dir = os.path.dirname(filename)
        self.filenames = {}
        self.parse_files()
        self.data = {}

        # Material properties/results
        self.input = self.initial_structure = self.lattice = self.final_structure = self.composition = None
        self.efermi = self.vbm = self.cbm = self.band_gap = None
        self.structures = []
        self.ionic_steps = []

        # parse the basic run parameters always
        self.parse_cp2k_params()
        self.parse_input()
        self.parse_global_params()
        self.parse_atomic_kind_info()
        self.parse_dft_params()
        self.parse_scf_params()

        # Auto-load will load the most crucial data into the data attribute
        if auto_load:
            self.ran_successfully()
            self.convergence()

            self.parse_structures()
            self.parse_energies()
            self.parse_forces()
            self.parse_stresses()
            self.parse_ionic_steps()

            self.parse_dos()
            self.parse_bandstructure()
            if not self.band_gap:
                self.parse_homo_lumo()
            if not self.vbm or not self.cbm:
                self.parse_mo_eigenvalues()
            self.parse_timing()

    @property
    def cp2k_version(self):
        """The cp2k version used in the calculation."""
        return self.data.get("cp2k_version")[0][0]

    @property
    def completed(self):
        """Did the calculation complete."""
        c = self.data.get("completed", False)
        if c:
            return c[0][0]
        return c

    @property
    def num_warnings(self):
        """How many warnings showed up during the run."""
        return self.data.get("num_warnings", 0)

    @property
    def run_type(self):
        """What type of run (Energy, MD, etc.) was performed."""
        return self.data.get("global").get("Run_type")

    @property
    def calculation_type(self):
        """Returns the calculation type (what io.vasp.outputs calls run_type)."""
        LDA_TYPES = {"LDA", "PADE", "BECKE88", "BECKE88_LR", "BECKE88_LR_ADIABATIC", "BECKE97"}

        GGA_TYPES = {"PBE", "PW92"}

        HYBRID_TYPES = {"BLYP", "B3LYP"}

        METAGGA_TYPES = {
            "TPSS": "TPSS",
            "RTPSS": "revTPSS",
            "M06L": "M06-L",
            "MBJ": "modified Becke-Johnson",
            "SCAN": "SCAN",
            "MS0": "MadeSimple0",
            "MS1": "MadeSimple1",
            "MS2": "MadeSimple2",
        }

        functional = self.data.get("dft", {}).get("functional", [None])
        ip = self.data.get("dft", {}).get("hfx", {}).get("Interaction_Potential")
        frac = self.data.get("dft", {}).get("hfx", {}).get("FRACTION")

        if len(functional) > 1:
            rt = "Mixed: " + ", ".join(functional)
            functional = " ".join(functional)
            if "HYP" in functional or (ip and frac) or (functional in HYBRID_TYPES):
                rt = "Hybrid"
        else:
            functional = functional[0]

            if functional is None:
                rt = "None"
            elif "HYP" in functional or (ip and frac) or (functional) in HYBRID_TYPES:
                rt = "Hybrid"
            elif "MGGA" in functional or functional in METAGGA_TYPES:
                rt = "METAGGA"
            elif "GGA" in functional or functional in GGA_TYPES:
                rt = "GGA"
            elif "LDA" in functional or functional in LDA_TYPES:
                rt = "LDA"
            else:
                rt = "Unknown"

        if self.is_hubbard:
            rt += "+U"
        if self.data.get("dft").get("vdw"):
            rt += "+VDW"

        return rt

    @property
    def project_name(self) -> str:
        """What project name was used for this calculation."""
        return self.data.get("global").get("project_name")

    @property
    def spin_polarized(self) -> bool:
        """Was the calculation spin polarized."""
        keys = ("UKS", "UNRESTRICTED_KOHN_SHAM", "LSD", "SPIN_POLARIZED")
        return any(key in self.data["dft"].values() for key in keys)

    @property
    def charge(self) -> float:
        """Get charge from the input file."""
        return self.input["FORCE_EVAL"]["DFT"].get("CHARGE", Keyword("", 0)).values[0]  # noqa: PD011

    @property
    def multiplicity(self) -> int:
        """Get the spin multiplicity from input file."""
        return self.input["FORCE_EVAL"]["DFT"].get("Multiplicity", Keyword("")).values[0]  # noqa: PD011

    @property
    def is_molecule(self) -> bool:
        """
        Returns True if the cp2k output was generated for a molecule (i.e.
        no periodicity in the cell). Returns false otherwise.
        """
        if self.data.get("poisson_periodicity", [[""]])[0][0].upper() == "NONE":
            return True
        return False

    @property
    def is_metal(self) -> bool:
        """Was a band gap found? i.e. is it a metal."""
        if self.band_gap is None:
            return True
        if self.band_gap <= 0:
            return True
        return False

    @property
    def is_hubbard(self) -> bool:
        """Returns True if hubbard +U correction was used."""
        for val in self.data.get("atomic_kind_info", {}).values():
            if val.get("DFT_PLUS_U", {}).get("U_MINUS_J", 0) > 0:
                return True
        return False

    def parse_files(self):
        """
        Identify files present in the directory with the cp2k output file. Looks for trajectories,
        dos, and cubes.
        """
        self.filenames["DOS"] = glob(os.path.join(self.dir, "*.dos*"))
        pdos = glob(os.path.join(self.dir, "*pdos*"))
        self.filenames["PDOS"] = []
        self.filenames["LDOS"] = []
        for p in pdos:
            if "list" in p.split("/")[-1]:
                self.filenames["LDOS"].append(p)
            else:
                self.filenames["PDOS"].append(p)
        self.filenames["band_structure"] = glob(os.path.join(self.dir, "*BAND.bs*"))
        self.filenames["trajectory"] = glob(os.path.join(self.dir, "*pos*.xyz*"))
        self.filenames["forces"] = glob(os.path.join(self.dir, "*frc*.xyz*"))
        self.filenames["stress"] = glob(os.path.join(self.dir, "*stress*"))
        self.filenames["cell"] = glob(os.path.join(self.dir, "*.cell*"))
        self.filenames["ener"] = glob(os.path.join(self.dir, "*.ener*"))
        self.filenames["electron_density"] = glob(os.path.join(self.dir, "*ELECTRON_DENSITY*.cube*"))
        self.filenames["spin_density"] = glob(os.path.join(self.dir, "*SPIN_DENSITY*.cube*"))
        self.filenames["v_hartree"] = glob(os.path.join(self.dir, "*hartree*.cube*"))
        self.filenames["hyperfine_tensor"] = glob(os.path.join(self.dir, "*HYPERFINE*eprhyp*"))
        self.filenames["g_tensor"] = glob(os.path.join(self.dir, "*GTENSOR*data*"))
        self.filenames["spinspin_tensor"] = glob(os.path.join(self.dir, "*K*data*"))
        self.filenames["chi_tensor"] = glob(os.path.join(self.dir, "*CHI*data*"))
        self.filenames["nmr_shift"] = glob(os.path.join(self.dir, "*SHIFT*data*"))
        self.filenames["raman"] = glob(os.path.join(self.dir, "*raman*data*"))
        restart = glob(os.path.join(self.dir, "*restart*"))
        self.filenames["restart.bak"] = []
        self.filenames["restart"] = []
        for r in restart:
            if "bak" in r.split("/")[-1]:
                self.filenames["restart.bak"].append(r)
            else:
                self.filenames["restart"].append(r)

        wfn = glob(os.path.join(self.dir, "*.wfn*")) + glob(os.path.join(self.dir, "*.kp*"))
        self.filenames["wfn.bak"] = []
        for w in wfn:
            if "bak" in w.split("/")[-1]:
                self.filenames["wfn.bak"].append(w)
            else:
                self.filenames["wfn"] = w
        for f in self.filenames.values():
            if hasattr(f, "sort"):
                f.sort(key=natural_keys)

    def parse_structures(self, trajectory_file=None, lattice_file=None):
        """
        Parses the structures from a cp2k calculation. Static calculations simply use the initial
        structure. For calculations with ionic motion, the function will look for the appropriate
        trajectory and lattice files based on naming convention. If no file is given, and no file
        is found, it is assumed that the lattice/structure remained constant, and the initial
        lattice/structure is used. Cp2k does not output the trajectory in the main output file by
        default, so non static calculations have to reference the trajectory file.
        """
        self.parse_initial_structure()
        trajectory_file = trajectory_file if trajectory_file else self.filenames.get("trajectory")
        if isinstance(trajectory_file, list):
            if len(trajectory_file) == 1:
                trajectory_file = trajectory_file[0]
            elif len(trajectory_file) > 1:
                raise FileNotFoundError("Unable to automatically determine trajectory file. More than one exist.")

        if lattice_file is None:
            if len(self.filenames["cell"]) == 0:
                lattices = self.parse_cell_params()
            elif len(self.filenames["cell"]) == 1:
                latt_file = np.loadtxt(self.filenames["cell"][0])
                lattices = (
                    [latt[2:11].reshape(3, 3) for latt in latt_file]
                    if len(latt_file.shape) > 1
                    else [latt_file[2:11].reshape(3, 3)]
                )
            else:
                raise FileNotFoundError("Unable to automatically determine lattice file. More than one exist.")
        else:
            latt_file = np.loadtxt(lattice_file)
            lattices = [latt[2:].reshape(3, 3) for latt in latt_file]

        if not trajectory_file:
            self.structures = []
            self.structures.append(self.initial_structure)
            self.final_structure = self.structures[-1]
        else:
            mols = XYZ.from_file(trajectory_file).all_molecules
            for mol in mols:
                mol.set_charge_and_spin(charge=self.charge, spin_multiplicity=self.multiplicity)
            self.structures = []
            gs = self.initial_structure.site_properties.get("ghost")
            if not self.is_molecule:
                for mol, latt in zip(mols, lattices):
                    self.structures.append(
                        Structure(
                            lattice=latt,
                            coords=[s.coords for s in mol.sites],
                            species=[s.specie for s in mol.sites],
                            coords_are_cartesian=True,
                            site_properties={"ghost": gs} if gs else {},
                            charge=self.charge,
                        )
                    )
            else:
                self.structures = mols
            self.final_structure = self.structures[-1]

    def parse_initial_structure(self):
        """Parse the initial structure from the main cp2k output file."""
        pattern = re.compile(r"- Atoms:\s+(\d+)")
        patterns = {"num_atoms": pattern}
        self.read_pattern(
            patterns=patterns,
            reverse=False,
            terminate_on_match=True,
            postprocess=int,
        )

        coord_table = []
        with zopen(self.filename, "rt") as f:
            while True:
                line = f.readline()
                if "Atom  Kind  Element       X           Y           Z          Z(eff)       Mass" in line:
                    for _ in range(self.data["num_atoms"][0][0]):
                        line = f.readline().split()
                        if line == []:
                            line = f.readline().split()
                        coord_table.append(line)
                    break

        lattice = self.parse_cell_params()
        gs = {}
        self.data["atomic_kind_list"] = []
        for v in self.data["atomic_kind_info"].values():
            if v["pseudo_potential"].upper() == "NONE":
                gs[v["kind_number"]] = True
            else:
                gs[v["kind_number"]] = False

        for c in coord_table:
            for k, v in self.data["atomic_kind_info"].items():
                if int(v["kind_number"]) == int(c[1]):
                    v["element"] = c[2]
                    self.data["atomic_kind_list"].append(k)
                    break

        if self.is_molecule:
            self.initial_structure = Molecule(
                species=[i[2] for i in coord_table],
                coords=[[float(i[4]), float(i[5]), float(i[6])] for i in coord_table],
                site_properties={"ghost": [gs.get(int(i[1])) for i in coord_table]},
                charge=self.charge,
                spin_multiplicity=self.multiplicity,
            )
        else:
            self.initial_structure = Structure(
                lattice,
                species=[i[2] for i in coord_table],
                coords=[[float(i[4]), float(i[5]), float(i[6])] for i in coord_table],
                coords_are_cartesian=True,
                site_properties={"ghost": [gs.get(int(i[1])) for i in coord_table]},
                charge=self.charge,
            )

        self.composition = self.initial_structure.composition
        return self.initial_structure

    def ran_successfully(self):
        """
        Sanity checks that the program ran successfully. Looks at the bottom of the CP2K output
        file for the "PROGRAM ENDED" line, which is printed when successfully ran. Also grabs
        the number of warnings issued.
        """
        program_ended_at = re.compile(r"PROGRAM ENDED AT\s+(\w+)")
        num_warnings = re.compile(r"The number of warnings for this run is : (\d+)")
        self.read_pattern(
            patterns={"completed": program_ended_at},
            reverse=True,
            terminate_on_match=True,
            postprocess=bool,
        )
        self.read_pattern(
            patterns={"num_warnings": num_warnings},
            reverse=True,
            terminate_on_match=True,
            postprocess=int,
        )

        if not self.completed:
            raise ValueError("The provided CP2K job did not finish running! Cannot parse the file reliably.")

    def convergence(self):
        """Check whether or not the SCF and geometry optimization cycles converged."""
        # SCF Loops
        uncoverged_inner_loop = re.compile(r"(Leaving inner SCF loop)")
        scf_converged = re.compile(r"(SCF run converged)|(SCF run NOT converged)")
        self.read_pattern(
            patterns={
                "uncoverged_inner_loop": uncoverged_inner_loop,
                "scf_converged": scf_converged,
            },
            reverse=True,
            terminate_on_match=False,
            postprocess=bool,
        )
        for i, x in enumerate(self.data["scf_converged"]):
            if x[0]:
                self.data["scf_converged"][i] = True
            else:
                self.data["scf_converged"][i] = False

        # GEO_OPT
        geo_opt_not_converged = re.compile(r"(MAXIMUM NUMBER OF OPTIMIZATION STEPS REACHED)")
        geo_opt_converged = re.compile(r"(GEOMETRY OPTIMIZATION COMPLETED)")
        self.read_pattern(
            patterns={
                "geo_opt_converged": geo_opt_converged,
                "geo_opt_not_converged": geo_opt_not_converged,
            },
            reverse=True,
            terminate_on_match=True,
            postprocess=bool,
        )

        if not all(self.data["scf_converged"]):
            warnings.warn(
                "There is at least one unconverged SCF cycle in the provided cp2k calculation",
                UserWarning,
            )
        if any(self.data["geo_opt_not_converged"]):
            warnings.warn("Geometry optimization did not converge", UserWarning)

    def parse_energies(self):
        """
        Get the total energy from a CP2K calculation. Presently, the energy reported in the
        trajectory (pos.xyz) file takes presidence over the energy reported in the main output
        file. This is because the trajectory file keeps track of energies in between restarts,
        while the main output file may or may not depending on whether a particular machine
        overwrites or appends it.
        """
        if self.filenames.get("trajectory"):
            toten_pattern = r".*E\s+\=\s+(-?\d+.\d+)"
            matches = regrep(
                self.filenames["trajectory"][-1],
                {"total_energy": toten_pattern},
                postprocess=float,
            )
            self.data["total_energy"] = list(
                chain.from_iterable(np.multiply([i[0] for i in matches.get("total_energy", [[]])], Ha_to_eV))
            )
        else:
            toten_pattern = re.compile(r"Total FORCE_EVAL.*\s(-?\d+.\d+)")
            self.read_pattern(
                {"total_energy": toten_pattern},
                terminate_on_match=False,
                postprocess=float,
                reverse=False,
            )
            self.data["total_energy"] = list(
                chain.from_iterable(np.multiply(self.data.get("total_energy", []), Ha_to_eV))
            )
        self.final_energy = self.data.get("total_energy", [])[-1]

    def parse_forces(self):
        """Get the forces from the forces file, or from the main output file."""
        if len(self.filenames["forces"]) == 1:
            self.data["forces"] = [
                [list(atom.coords) for atom in step]
                for step in XYZ.from_file(self.filenames["forces"][0]).all_molecules
            ]
        else:
            header_pattern = r"ATOMIC FORCES.+Z"
            row_pattern = r"\s+\d+\s+\d+\s+\w+\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)"
            footer_pattern = r"SUM OF ATOMIC FORCES"

            self.data["forces"] = self.read_table_pattern(
                header_pattern=header_pattern,
                row_pattern=row_pattern,
                footer_pattern=footer_pattern,
                postprocess=postprocessor,
                last_one_only=False,
            )

    def parse_stresses(self):
        """Get the stresses from stress file, or from the main output file."""
        if len(self.filenames["stress"]) == 1:
            dat = np.genfromtxt(self.filenames["stress"][0], skip_header=1)
            dat = [dat] if len(np.shape(dat)) == 1 else dat
            self.data["stress_tensor"] = [[list(d[2:5]), list(d[5:8]), list(d[8:11])] for d in dat]
        else:
            header_pattern = r"STRESS\|\s+x\s+y\s+z"
            row_pattern = (
                r"STRESS\|\s+[?:x|y|z]\s+(-?\d+\.\d+E?[-|\+]?\d+)\s+"
                r"(-?\d+\.\d+E?[-|\+]?\d+)\s+(-?\d+\.\d+E?[-|\+]?\d+).*$"
            )
            footer_pattern = r"^$"
            d = self.read_table_pattern(
                header_pattern=header_pattern,
                row_pattern=row_pattern,
                footer_pattern=footer_pattern,
                postprocess=postprocessor,
                last_one_only=False,
            )

            def chunks(lst, n):
                """Yield successive n-sized chunks from lst."""
                for i in range(0, len(lst), n):
                    if i % 2 == 0:
                        yield lst[i : i + n]

            if d:
                self.data["stress_tensor"] = list(chunks(d[0], 3))

    def parse_ionic_steps(self):
        """Parse the ionic step info. If already parsed, this will just assimilate."""
        if not self.structures:
            self.parse_structures()
        if not self.data.get("total_energy"):
            self.parse_energies()
        if not self.data.get("forces"):
            self.parse_forces()
        if not self.data.get("stress_tensor"):
            self.parse_stresses()

        for i, (structure, energy) in enumerate(zip(self.structures, self.data.get("total_energy"))):
            self.ionic_steps.append(
                {
                    "structure": structure,
                    "E": energy,
                    "forces": self.data["forces"][i] if self.data.get("forces") else None,
                    "stress_tensor": self.data["stress_tensor"][i] if self.data.get("stress_tensor") else None,
                }
            )

        return self.ionic_steps

    def parse_cp2k_params(self):
        """Parse the CP2K general parameters from CP2K output file into a dictionary."""
        version = re.compile(r"\s+CP2K\|.+version\s+(.+)")
        input_file = re.compile(r"\s+CP2K\|\s+Input file name\s+(.+)$")
        self.read_pattern(
            {"cp2k_version": version, "input_filename": input_file},
            terminate_on_match=True,
            reverse=False,
            postprocess=str,
        )

    def parse_plus_u_params(self):
        """Parse the DFT+U params."""
        method = re.compile(r"\s+DFT\+U\|\s+Method\s+()$")
        self.read_pattern(
            {"dft_plus_u_method": method},
            terminate_on_match=True,
            reverse=False,
            postprocess=postprocessor,
        )

    def parse_input(self):
        """Load in the input set from the input file (if it can be found)."""
        if len(self.data["input_filename"]) == 0:
            return
        input_filename = self.data["input_filename"][0][0]
        for ext in ["", ".gz", ".GZ", ".z", ".Z", ".bz2", ".BZ2"]:
            if os.path.exists(os.path.join(self.dir, input_filename + ext)):
                self.input = Cp2kInput.from_file(os.path.join(self.dir, input_filename + ext))
                return
        warnings.warn("Original input file not found. Some info may be lost.")

    def parse_global_params(self):
        """Parse the GLOBAL section parameters from CP2K output file into a dictionary."""
        pat = re.compile(r"\s+GLOBAL\|\s+([\w+\s]*)\s+(\w+)")
        self.read_pattern({"global": pat}, terminate_on_match=False, reverse=False)
        for d in self.data["global"]:
            d[0], d[1] = postprocessor(d[0]), str(d[1])
        self.data["global"] = dict(self.data["global"])

    def parse_dft_params(self):
        """Parse the DFT parameters (as well as functional, HF, vdW params)."""
        pat = re.compile(r"\s+DFT\|\s+(\w.*)\s\s\s(.*)$")
        self.read_pattern(
            {"dft": pat},
            terminate_on_match=False,
            postprocess=postprocessor,
            reverse=False,
        )
        self.data["dft"] = dict(self.data["dft"])

        self.data["dft"]["cutoffs"] = {}
        self.data["dft"]["cutoffs"]["density"] = self.data["dft"].pop("Cutoffs:_density", None)
        self.data["dft"]["cutoffs"]["gradient"] = self.data["dft"].pop("gradient", None)
        self.data["dft"]["cutoffs"]["tau"] = self.data["dft"].pop("tau", None)

        # Functional
        if self.input and self.input.check("FORCE_EVAL/DFT/XC/XC_FUNCTIONAL"):
            xc_funcs = list(self.input["force_eval"]["dft"]["xc"]["xc_functional"].subsections)
            if xc_funcs:
                self.data["dft"]["functional"] = xc_funcs
            else:
                for v in self.input["force_eval"]["dft"]["xc"].subsections.values():
                    if v.name.upper() == "XC_FUNCTIONAL":
                        self.data["dft"]["functional"] = v.section_parameters
        else:
            functional = re.compile(r"\s+FUNCTIONAL\|\s+(.+):")
            self.read_pattern(
                {"functional": functional},
                terminate_on_match=False,
                postprocess=postprocessor,
                reverse=False,
            )
            self.data["dft"]["functional"] = [item for sublist in self.data.pop("functional", None) for item in sublist]

        # DFT+U
        self.data["dft"]["dft_plus_u"] = self.is_hubbard

        # HF exchange info
        hfx = re.compile(r"\s+HFX_INFO\|\s+(.+):\s+(.*)$")
        self.read_pattern({"hfx": hfx}, terminate_on_match=False, postprocess=postprocessor, reverse=False)
        self.data["dft"]["hfx"] = dict(self.data.pop("hfx"))

        # Van der waals correction
        vdw = re.compile(r"\s+vdW POTENTIAL\|(.+)$")
        self.read_pattern(
            {"vdw": vdw},
            terminate_on_match=False,
            postprocess=lambda x: str(x).strip(),
            reverse=False,
        )
        if self.data.get("vdw"):
            found = False
            suffix = ""
            for ll in self.data.get("vdw"):
                for _possible, _name in zip(
                    ["RVV10", "LMKLL", "DRSLL", "DFT-D3", "DFT-D2"],
                    ["RVV10", "LMKLL", "DRSLL", "D3", "D2"],
                ):
                    if _possible in ll[0]:
                        found = _name
                    if "BJ" in ll[0]:
                        suffix = "(BJ)"

            self.data["dft"]["vdw"] = found + suffix if found else self.data.pop("vdw")[0][0]

        poisson_periodic = {"poisson_periodicity": re.compile(r"POISSON\| Periodicity\s+(\w+)")}
        self.read_pattern(poisson_periodic, terminate_on_match=True)

    def parse_qs_params(self):
        """Parse the DFT parameters (as well as functional, HF, vdW params)."""
        pat = re.compile(r"\s+QS\|\s+(\w.*)\s\s\s(.*)$")
        self.read_pattern(
            {"QS": pat},
            terminate_on_match=False,
            postprocess=postprocessor,
            reverse=False,
        )
        self.data["QS"] = dict(self.data["QS"])
        tmp = {}
        i = 1
        for k in list(self.data["QS"]):
            if "grid_level" in str(k) and "Number" not in str(k):
                tmp[i] = self.data["QS"].pop(k)
                i += 1
        self.data["QS"]["Multi_grid_cutoffs_[a.u.]"] = tmp

    def parse_overlap_condition(self):
        """
        Retrieve the overlap condition number
        """
        overlap_condition = re.compile(r"\|A\|\*\|A\^-1\|.+=\s+(-?\d+\.\d+E[+\-]?\d+)\s+Log")
        self.read_pattern(
            {"overlap_condition_number": overlap_condition},
            terminate_on_match=True,
            reverse=False,
            postprocess=postprocessor,
        )

    def parse_scf_params(self):
        """
        Retrieve the most import SCF parameters: the max number of scf cycles (max_scf),
        the convergence cutoff for scf (eps_scf),.
        """
        max_scf = re.compile(r"max_scf:\s+(\d+)")
        eps_scf = re.compile(r"eps_scf:\s+(\d+)")
        self.read_pattern(
            {"max_scf": max_scf, "eps_scf": eps_scf},
            terminate_on_match=True,
            reverse=False,
        )
        self.data["scf"] = {}
        self.data["scf"]["max_scf"] = self.data.pop("max_scf")[0][0] if self.data["max_scf"] else None
        self.data["scf"]["eps_scf"] = self.data.pop("eps_scf")[0][0] if self.data["eps_scf"] else None

    def parse_cell_params(self):
        """Parse the lattice parameters (initial) from the output file."""
        if self.input and self.input.check("force_eval/subsys/cell"):
            cell = self.input["force_eval"]["subsys"]["cell"]
            if cell.get("abc"):
                return [
                    [cell["abc"].values[0], 0, 0],  # noqa: PD011
                    [0, cell["abc"].values[1], 0],  # noqa: PD011
                    [0, 0, cell["abc"].values[2]],  # noqa: PD011
                ]
            return [
                list(cell.get("A").values),
                list(cell.get("B").values),
                list(cell.get("C").values),
            ]

        warnings.warn(
            "Input file lost. Reading cell params from summary at top of output. Precision errors may result."
        )
        cell_volume = re.compile(r"\s+CELL\|\sVolume.*\s(\d+\.\d+)")
        vectors = re.compile(r"\s+CELL\| Vector.*\s(-?\d+\.\d+)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)")
        angles = re.compile(r"\s+CELL\| Angle.*\s(\d+\.\d+)")

        self.read_pattern(
            {"cell_volume": cell_volume, "lattice": vectors, "angles": angles},
            terminate_on_match=False,
            postprocess=float,
            reverse=False,
        )
        i = iter(self.data["lattice"])
        lattices = list(zip(i, i, i))
        return lattices[0]

    def parse_atomic_kind_info(self):
        """
        Parse info on what atomic kinds are present and what basis/pseudopotential is describing
        each of them.
        """
        kinds = re.compile(r"Atomic kind: (\w+)")
        orbital_basis_set = re.compile(r"Orbital Basis Set\s+(.+$)")
        potential_information = re.compile(r"(?:Potential information for\s+(.+$))|(?:atomic kind are GHOST atoms)")
        auxiliary_basis_set = re.compile(r"Auxiliary Fit Basis Set\s+(.+$)")
        core_electrons = re.compile(r"Total number of core electrons\s+(\d+)")
        valence_electrons = re.compile(r"Total number of valence electrons\s+(\d+)")
        pseudo_energy = re.compile(r"Total Pseudopotential Energy.+(-?\d+.\d+)")
        self.read_pattern(
            {
                "kinds": kinds,
                "orbital_basis_set": orbital_basis_set,
                "potential_info": potential_information,
                "auxiliary_basis_set": auxiliary_basis_set,
                "core_electrons": core_electrons,
                "valence_electrons": valence_electrons,
                "pseudo_energy": pseudo_energy,
            },
            terminate_on_match=True,
            postprocess=str,
            reverse=False,
        )
        atomic_kind_info = {}
        _kinds = []
        for _ in list(chain.from_iterable(self.data["kinds"])):
            if _ not in _kinds:
                _kinds.append(_)
        for i, kind in enumerate(_kinds):
            atomic_kind_info[kind] = {
                "orbital_basis_set": self.data.get("orbital_basis_set")[i][0],
                "pseudo_potential": self.data.get("potential_info")[i][0],
                "kind_number": i + 1,
            }
            try:
                if self.data.get("valence_electrons"):
                    tmp = self.data.get("valence_electrons")[i][0]
                elif self.data.get("potential_info")[i][0].upper() == "NONE":
                    tmp = 0
                else:
                    tmp = self.data.get("potential_info")[i][0].split("q")[-1]
                atomic_kind_info[kind]["valence_electrons"] = int(tmp)
            except (TypeError, IndexError, ValueError):
                atomic_kind_info[kind]["valence_electrons"] = None
            try:
                atomic_kind_info[kind]["core_electrons"] = int(self.data.get("core_electrons")[i][0])
            except (TypeError, IndexError, ValueError):
                atomic_kind_info[kind]["core_electrons"] = None
            try:
                atomic_kind_info[kind]["auxiliary_basis_set"] = self.data.get("auxiliary_basis_set")[i]
            except (TypeError, IndexError):
                atomic_kind_info[kind]["auxiliary_basis_set"] = None
            try:
                atomic_kind_info[kind]["total_pseudopotential_energy"] = float(
                    self.data.get("total_pseudopotential_energy")[i][0] * Ha_to_eV
                )
            except (TypeError, IndexError, ValueError):
                atomic_kind_info[kind]["total_pseudopotential_energy"] = None

        with zopen(self.filename, "rt") as f:
            j = -1
            lines = f.readlines()
            for k, line in enumerate(lines):
                if "MOLECULE KIND INFORMATION" in line:
                    break
                if "Atomic kind" in line:
                    j += 1
                if "DFT+U correction" in line:
                    atomic_kind_info[_kinds[j]]["DFT_PLUS_U"] = {
                        "L": int(lines[k + 1].split()[-1]),
                        "U_MINUS_J": float(lines[k + 2].split()[-1]),
                    }
                k += 1

        self.data["atomic_kind_info"] = atomic_kind_info

    def parse_total_numbers(self):
        """Parse total numbers (not usually important)."""
        atomic_kinds = r"- Atomic kinds:\s+(\d+)"
        atoms = r"- Atoms:\s+(\d+)"
        shell_sets = r"- Shell sets:\s+(\d+)"
        shells = r"- Shells:\s+(\d+)"
        primitive_funcs = r"- Primitive Cartesian functions:\s+(\d+)"
        cart_base_funcs = r"- Cartesian basis functions:\s+(\d+)"
        spher_base_funcs = r"- Spherical basis functions:\s+(\d+)"

        self.read_pattern(
            {
                "atomic_kinds": atomic_kinds,
                "atoms": atoms,
                "shell_sets": shell_sets,
                "shells": shells,
                "primitive_cartesian_functions": primitive_funcs,
                "cartesian_basis_functions": cart_base_funcs,
                "spherical_basis_functions": spher_base_funcs,
            },
            terminate_on_match=True,
        )

    def parse_scf_opt(self):
        """Parse the SCF cycles (not usually important)."""
        header = r"Step\s+Update method\s+Time\s+Convergence\s+Total energy\s+Change\s+\-+"
        row = (
            r"(\d+)"
            r"\s+([A-Za-z\./_]+\s?[A-Za-z\./]+)"
            r"\s+(-?\d+\.\d+(?:[eE][+\-]?\d+)?)"
            r"\s+(-?\d+\.\d+(?:[eE][+\-]?\d+)?)"
            r"(\s+-?\d+\.\d+(?:[eE][+\-]?\d+)?)?"
            r"\s+(-?\d+\.\d+(?:[eE][+\-]?\d+)?)"
            r"(\s+-?\d+\.\d+(?:[eE][+\-]?\d+)?)?"
        )

        footer = r"^$"

        scfs = self.read_table_pattern(
            header_pattern=header,
            row_pattern=row,
            footer_pattern=footer,
            last_one_only=False,
            strip=["HFX_MEM_INFO", "*"],
        )

        self.data["electronic_steps"] = []
        self.data["convergence"] = []
        self.data["scf_time"] = []
        for i in scfs:
            self.data["scf_time"].append([float(j[-4]) for j in i])
            self.data["convergence"].append([float(j[-3]) for j in i if j[-3] != "None"])
            self.data["electronic_steps"].append([float(j[-2]) for j in i])

    def parse_timing(self):
        """Parse the timing info (how long did the run take)."""
        header = r"SUBROUTINE\s+CALLS\s+ASD\s+SELF TIME\s+TOTAL TIME\s+MAXIMUM\s+AVERAGE\s+MAXIMUM\s+AVERAGE\s+MAXIMUM"
        row = r"(\w+)\s+(.+)\s+(\d+\.\d+)\s+(\d+\.\d+)\s+(\d+\.\d+)\s+(\d+\.\d+)\s+(\d+\.\d+)"
        footer = r"\-+"

        timing = self.read_table_pattern(
            header_pattern=header,
            row_pattern=row,
            footer_pattern=footer,
            last_one_only=True,
            postprocess=postprocessor,
        )
        self.timing = {}
        for t in timing:
            self.timing[t[0]] = {
                "calls": {"max": t[1]},
                "asd": t[2],
                "self_time": {"average": t[3], "maximum": t[4]},
                "total_time": {"average": t[5], "maximum": t[6]},
            }

    def parse_opt_steps(self):
        """Parse the geometry optimization information."""
        # "Information at step =" Summary block (floating point terms)
        total_energy = re.compile(r"\s+Total Energy\s+=\s+(-?\d+.\d+)")
        real_energy_change = re.compile(r"\s+Real energy change\s+=\s+(-?\d+.\d+)")
        prediced_change_in_energy = re.compile(r"\s+Predicted change in energy\s+=\s+(-?\d+.\d+)")
        scaling_factor = re.compile(r"\s+Scaling factor\s+=\s+(-?\d+.\d+)")
        step_size = re.compile(r"\s+Step size\s+=\s+(-?\d+.\d+)")
        trust_radius = re.compile(r"\s+Trust radius\s+=\s+(-?\d+.\d+)")
        used_time = re.compile(r"\s+Used time\s+=\s+(-?\d+.\d+)")

        # For RUN_TYPE=CELL_OPT
        pressure_deviation = re.compile(r"\s+Pressure Deviation.*=\s+(-?\d+.\d+)")
        pressure_tolerance = re.compile(r"\s+Pressure Tolerance.*=\s+(-?\d+.\d+)")

        self.read_pattern(
            {
                "total_energy": total_energy,
                "real_energy_change": real_energy_change,
                "predicted_change_in_energy": prediced_change_in_energy,
                "scaling_factor": scaling_factor,
                "step_size": step_size,
                "trust_radius": trust_radius,
                "used_time": used_time,
                "pressure_deviation": pressure_deviation,
                "pressure_tolerance": pressure_tolerance,
            },
            terminate_on_match=False,
            postprocess=float,
        )

        # "Information at step =" Summary block (bool terms)
        decrease_in_energy = re.compile(r"\s+Decrease in energy\s+=\s+(\w+)")
        converged_step_size = re.compile(r"\s+Convergence in step size\s+=\s+(\w+)")
        converged_rms_step = re.compile(r"\s+Convergence in RMS step\s+=\s+(\w+)")
        converged_in_grad = re.compile(r"\s+Conv\. in gradients\s+=\s+(\w+)")
        converged_in_rms_grad = re.compile(r"\s+Conv\. in RMS gradients\s+=\s+(\w+)")
        pressure_converged = re.compile(r"\s+Conv\. for  PRESSURE\s+=\s+(\w+)")

        self.read_pattern(
            {
                "decrease_in_energy": decrease_in_energy,
                "converged_step_size": converged_step_size,
                "converged_rms_step": converged_rms_step,
                "converged_in_grad": converged_in_grad,
                "converged_in_rms_grad": converged_in_rms_grad,
                "pressure_converged": pressure_converged,
            },
            terminate_on_match=False,
            postprocess=postprocessor,
        )

    def parse_mulliken(self):
        """
        Parse the mulliken population analysis info for each step
        """
        header = r"Mulliken Population Analysis.+Net charge"
        pattern = r"\s+(\d)\s+(\w+)\s+(\d+)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)"
        footer = r".+Total charge"

        d = self.read_table_pattern(
            header_pattern=header,
            row_pattern=pattern,
            footer_pattern=footer,
            last_one_only=False,
        )
        if d:
            print("Found data, but not yet implemented!")

    def parse_hirshfeld(self):
        """Parse the hirshfeld population analysis for each step."""
        uks = self.spin_polarized
        header = r"Hirshfeld Charges.+Net charge"
        footer = r"^$"

        if not uks:
            pattern = r"\s+(\d)\s+(\w+)\s+(\d+)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)"
            d = self.read_table_pattern(
                header_pattern=header,
                row_pattern=pattern,
                footer_pattern=footer,
                last_one_only=False,
            )
            for i, ionic_step in enumerate(d):
                population = []
                net_charge = []
                for site in ionic_step:
                    population.append(site[4])
                    net_charge.append(site[5])
                hirshfeld = [{"population": population[j], "net_charge": net_charge[j]} for j in range(len(population))]
                self.structures[i].add_site_property("hirshfield", hirshfeld)
        else:
            pattern = (
                r"\s+(\d)\s+(\w+)\s+(\d+)\s+(-?\d+\.\d+)\s+"
                r"(-?\d+\.\d+)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)"
            )
            d = self.read_table_pattern(
                header_pattern=header,
                row_pattern=pattern,
                footer_pattern=footer,
                last_one_only=False,
            )
            for i, ionic_step in enumerate(d):
                population = []
                net_charge = []
                spin_moment = []
                for site in ionic_step:
                    population.append(tuple(site[4:5]))
                    spin_moment.append(site[6])
                    net_charge.append(site[7])
                hirshfeld = [
                    {
                        "population": population[j],
                        "net_charge": net_charge[j],
                        "spin_moment": spin_moment[j],
                    }
                    for j in range(len(population))
                ]
                self.structures[i].add_site_property("hirshfield", hirshfeld)

    def parse_mo_eigenvalues(self):
        """
        Parse the MO eigenvalues from the cp2k output file. Will get the eigenvalues (and band gap)
        at each ionic step (if more than one exist).

        Everything is decomposed by spin channel. If calculation was performed without spin
        polarization, then only Spin.up will be present, which represents the average of up and
        down.
        """
        eigenvalues = []
        efermi = []

        with zopen(self.filename, "rt") as f:
            lines = iter(f.readlines())
            for line in lines:
                try:
                    if " occupied subspace spin" in line:
                        eigenvalues.append(
                            {
                                "occupied": {Spin.up: [], Spin.down: []},
                                "unoccupied": {Spin.up: [], Spin.down: []},
                            }
                        )
                        efermi.append({Spin.up: np.nan, Spin.down: np.nan})
                        next(lines)
                        while True:
                            line = next(lines)
                            if "Fermi" in line:
                                efermi[-1][Spin.up] = float(line.split()[-1])
                                break
                            eigenvalues[-1]["occupied"][Spin.up].extend([Ha_to_eV * float(val) for val in line.split()])
                        next(lines)
                        line = next(lines)
                        if " occupied subspace spin" in line:
                            next(lines)
                            while True:
                                line = next(lines)
                                if "Fermi" in line:
                                    efermi[-1][Spin.down] = float(line.split()[-1])
                                    break
                                eigenvalues[-1]["occupied"][Spin.down].extend(
                                    [Ha_to_eV * float(val) for val in line.split()]
                                )
                    if " unoccupied subspace spin" in line:
                        next(lines)
                        line = next(lines)
                        while True:
                            if "WARNING : did not converge" in line:
                                warnings.warn(
                                    "Convergence of eigenvalues for unoccupied subspace spin 1 did NOT converge"
                                )
                                next(lines)
                                next(lines)
                                next(lines)
                                line = next(lines)
                                eigenvalues[-1]["unoccupied"][Spin.up].extend(
                                    [Ha_to_eV * float(line) for line in line.split()]
                                )
                                next(lines)
                                line = next(lines)
                                break

                            if "convergence" in line:
                                line = next(lines)

                            if "eigenvalues" in line.lower() or "HOMO" in line or "|" in line:
                                break
                            eigenvalues[-1]["unoccupied"][Spin.up].extend(
                                [Ha_to_eV * float(val) for val in line.split()]
                            )
                            line = next(lines)

                        if " unoccupied subspace spin" in line:
                            next(lines)
                            line = next(lines)
                            while True:
                                if "WARNING : did not converge" in line:
                                    warnings.warn(
                                        "Convergence of eigenvalues for unoccupied subspace spin 2 did NOT converge"
                                    )
                                    next(lines)
                                    next(lines)
                                    next(lines)
                                    line = next(lines)
                                    eigenvalues[-1]["unoccupied"][Spin.down].extend(
                                        [Ha_to_eV * float(line) for line in line.split()]
                                    )
                                    break

                                if "convergence" in line:
                                    line = next(lines)

                                if "HOMO" in line or "|" in line:
                                    next(lines)
                                    break
                                try:
                                    eigenvalues[-1]["unoccupied"][Spin.down].extend(
                                        [Ha_to_eV * float(val) for val in line.split()]
                                    )
                                except AttributeError:
                                    break
                                line = next(lines)

                except ValueError:
                    eigenvalues = [
                        {
                            "occupied": {Spin.up: None, Spin.down: None},
                            "unoccupied": {Spin.up: None, Spin.down: None},
                        }
                    ]
                    warnings.warn("Convergence of eigenvalues for one or more subspaces did NOT converge")

        self.data["eigenvalues"] = eigenvalues

        if len(eigenvalues) == 0:
            return

        if self.spin_polarized:
            self.data["vbm"] = {
                Spin.up: np.max(eigenvalues[-1]["occupied"][Spin.up]),
                Spin.down: np.max(eigenvalues[-1]["occupied"][Spin.down]),
            }
            self.data["cbm"] = {
                Spin.up: np.nanmin(eigenvalues[-1]["unoccupied"][Spin.up] or np.nan),
                Spin.down: np.nanmin(eigenvalues[-1]["unoccupied"][Spin.down] or np.nan),
            }
            self.vbm = np.nanmean(list(self.data["vbm"].values()))
            self.cbm = np.nanmean(list(self.data["cbm"].values()))
            self.efermi = np.nanmean(list(efermi[-1].values()))
        else:
            self.data["vbm"] = {
                Spin.up: np.max(eigenvalues[-1]["occupied"][Spin.up]),
                Spin.down: None,
            }
            self.data["cbm"] = {
                Spin.up: np.min(eigenvalues[-1]["unoccupied"][Spin.up]),
                Spin.down: None,
            }
            self.vbm = self.data["vbm"][Spin.up]
            self.cbm = self.data["cbm"][Spin.up]
            self.efermi = efermi[-1][Spin.up]

        num_occ = len(eigenvalues[-1]["occupied"][Spin.up])
        num_unocc = len(eigenvalues[-1]["unoccupied"][Spin.up])
        self.data["tdos"] = Dos(
            efermi=self.vbm + 1e-6,
            energies=list(eigenvalues[-1]["occupied"][Spin.up]) + list(eigenvalues[-1]["unoccupied"][Spin.down]),
            densities={
                Spin.up: [1 for _ in range(num_occ)] + [0 for _ in range(num_unocc)],
                Spin.down: [1 for _ in range(num_occ)] + [0 for _ in range(num_unocc)],
            },
        )

    def parse_homo_lumo(self):
        """
        Find the HOMO - LUMO gap in [eV]. Returns the last value. For gaps/eigenvalues decomposed
        by spin up/spin down channel and over many ionic steps, see parse_mo_eigenvalues().
        """
        pattern = re.compile(r"HOMO.*-.*LUMO.*gap.*\s(-?\d+.\d+)")
        self.read_pattern(
            patterns={"band_gap": pattern},
            reverse=True,
            terminate_on_match=False,
            postprocess=float,
        )
        bg = {Spin.up: [], Spin.down: []}
        for i in range(len(self.data["band_gap"])):
            if self.spin_polarized:
                if i % 2:
                    bg[Spin.up].append(self.data["band_gap"][i][0])
                else:
                    bg[Spin.down].append(self.data["band_gap"][i][0])
            else:
                bg[Spin.up].append(self.data["band_gap"][i][0])
                bg[Spin.down].append(self.data["band_gap"][i][0])
        self.data["band_gap"] = bg
        self.band_gap = (bg[Spin.up][-1] + bg[Spin.down][-1]) / 2 if bg[Spin.up] and bg[Spin.down] else None

    def parse_dos(self, dos_file=None, pdos_files=None, ldos_files=None):
        """
        Parse the dos files produced by cp2k calculation. CP2K produces different files based
        on the input file rather than assimilating them all into one file.

        One file type is the overall DOS file, which is used for k-point calculations. For
        non-kpoint calculation, the overall DOS is generally not calculated, but the
        element-projected pDOS is. Separate files are created for each spin channel and each
        atom kind. If requested, cp2k can also do site/local projected dos (ldos). Each site
        requested will have a separate file for each spin channel (if spin polarized calculation
        is performed).

        If possible, this function will assimilate the ldos files into a CompleteDos object.
        Either provide a list of PDOS file paths, or use glob to find the .pdos_ALPHA extension
        in the calculation directory.

        Args:
            dos_file (str): Name of the dos file, otherwise will be inferred
            pdos_files (list): list of pdos file paths, otherwise they will be inferred
            ldos_files (list): list of ldos file paths, otherwise they will be inferred
        """
        if dos_file is None:
            dos_file = self.filenames["DOS"][0] if self.filenames["DOS"] else None

        if pdos_files is None:
            pdos_files = self.filenames["PDOS"]

        if ldos_files is None:
            ldos_files = self.filenames["LDOS"]

        tdos, pdoss, ldoss = None, {}, {}
        # Parse specie projected dos
        for pdos_file in pdos_files:
            _pdos, _tdos = parse_pdos(pdos_file, total=True)
            for k in _pdos:
                if k in pdoss:
                    for orbital in _pdos[k]:
                        pdoss[k][orbital].densities.update(_pdos[k][orbital].densities)
                else:
                    pdoss.update(_pdos)
            if not tdos:
                tdos = _tdos
            else:
                for k, v in _tdos.densities.copy().items():
                    if k not in tdos.densities:
                        tdos.densities[Spin(int(k))] = [0] * len(v)
                    tdos.densities[k] = np.array(tdos.densities[k]) + np.array(_tdos.densities[k])

        # parse any site-projected dos
        for ldos_file in ldos_files:
            _pdos = parse_pdos(ldos_file)
            for k in _pdos:
                if k in ldoss:
                    for orbital in _pdos[k]:
                        ldoss[k][orbital].densities.update(_pdos[k][orbital].densities)
                else:
                    ldoss.update(_pdos)

        self.data["pdos"] = jsanitize(pdoss, strict=True)
        self.data["ldos"] = jsanitize(ldoss, strict=True)

        if dos_file:
            self.data["tdos"] = parse_dos(dos_file)
        else:
            self.data["tdos"] = tdos

        if self.data.get("tdos"):
            self.band_gap = self.data["tdos"].get_gap()
            self.cbm, self.vbm = self.data["tdos"].get_cbm_vbm()
            self.efermi = (self.cbm + self.vbm) / 2

        # If number of site-projected dos == number of sites, assume they are bijective
        # and create the CompleteDos object
        _ldoss = {}
        if self.initial_structure and len(ldoss) == len(self.initial_structure):
            for k, lds in ldoss.items():
                _ldoss[self.initial_structure[int(k) - 1]] = {Orbital(orb): lds[orb].densities for orb in lds}

            self.data["cdos"] = CompleteDos(self.final_structure, total_dos=tdos, pdoss=_ldoss)

    @property
    def complete_dos(self) -> CompleteDos:
        """Returns complete dos object if it has been parsed."""
        return self.data.get("cdos")

    @property
    def band_structure(self) -> BandStructure:
        """Returns band structure object if it has been parsed."""
        return self.data.get("band_structure")

    def parse_bandstructure(self, bandstructure_filename=None) -> None:
        """
        Parse a CP2K bandstructure file.

        Args:
            bandstructure_filename: Filename containing bandstructure info. If
            not provided, then the pmg name of "BAND.bs" will be assumed by
            the filename parser.
        """
        if not bandstructure_filename:
            if self.filenames["band_structure"]:
                bandstructure_filename = self.filenames["band_structure"][0]
            else:
                return

        with open(bandstructure_filename) as f:
            lines = f.read().split("\n")

        data = np.loadtxt(bandstructure_filename)
        nkpts = int(lines[0].split()[6])
        nbands = int(lines[0].split()[-2])
        rec_lat = (
            self.final_structure.lattice.reciprocal_lattice
            if self.final_structure
            else self.initial_structure.lattice.reciprocal_lattice
        )

        labels = {}
        kpts = []
        nkpts = 0
        for line in lines:
            if not line.startswith("#"):
                continue
            if line.split()[1] == "Set":
                nkpts += int(lines[0].split()[6])
            elif line.split()[1] == "Point":
                kpts.append(list(map(float, line.split()[-4:-1])))
            elif line.split()[1] == "Special" in line:
                splt = line.split()
                label = splt[7]
                if label.upper() == "GAMMA":
                    label = "\\Gamma"
                kpt = np.array(splt[4:7]).astype(float).tolist()
                if label.upper() != "NONE":
                    labels[label] = kpt

        if self.spin_polarized:
            kpts = kpts[::2]

        eigenvals = {}
        if self.spin_polarized:
            up = data.reshape(-1, nbands * 2, data.shape[1])[:, :nbands].reshape(-1, data.shape[1])
            down = data.reshape(-1, nbands * 2, data.shape[1])[:, nbands:].reshape(-1, data.shape[1])
            eigenvals = {
                Spin.up: up[:, 1].reshape((nkpts, nbands)).T.tolist(),
                Spin.down: down[:, 1].reshape((nkpts, nbands)).T.tolist(),
            }
        else:
            eigenvals = {Spin.up: data.reshape((nbands, nkpts))}

        occ = data[:, 1][data[:, -1] != 0.0]
        homo = np.max(occ)
        unocc = data[:, 1][data[:, -1] == 0.0]
        lumo = np.min(unocc)
        efermi = (lumo + homo) / 2
        self.efermi = efermi

        self.data["band_structure"] = BandStructureSymmLine(
            kpoints=kpts,
            eigenvals=eigenvals,
            lattice=rec_lat,
            efermi=efermi,
            labels_dict=labels,
            structure=self.final_structure,
            projections=None,  # not implemented in cp2k
        )

        self.band_gap = self.data["band_structure"].get_band_gap().get("energy")
        self.vbm = self.data["band_structure"].get_vbm().get("energy")
        self.cbm = self.data["band_structure"].get_cbm().get("energy")

    def parse_hyperfine(self, hyperfine_filename=None):
        """Parse a file containing hyperfine coupling tensors for each atomic site."""
        if not hyperfine_filename:
            if self.filenames["hyperfine_tensor"]:
                hyperfine_filename = self.filenames["hyperfine_tensor"][0]
            else:
                return None

        with zopen(hyperfine_filename, "rt") as f:
            lines = [line for line in f.read().split("\n") if line]

        hyperfine = [[] for _ in self.ionic_steps]
        for i in range(2, len(lines), 5):
            x = list(map(float, lines[i + 2].split()[-3:]))
            y = list(map(float, lines[i + 3].split()[-3:]))
            z = list(map(float, lines[i + 4].split()[-3:]))
            hyperfine[-1].append([x, y, z])

        self.data["hyperfine_tensor"] = hyperfine
        return hyperfine

    def parse_gtensor(self, gtensor_filename=None):
        """Parse a file containing g tensor."""
        if not gtensor_filename:
            if self.filenames["g_tensor"]:
                gtensor_filename = self.filenames["g_tensor"][0]
            else:
                return None

        with zopen(gtensor_filename, "rt") as f:
            lines = [line for line in f.read().split("\n") if line]

        data = {}
        data["gmatrix_zke"] = []
        data["gmatrix_so"] = []
        data["gmatrix_soo"] = []
        data["gmatrix_total"] = []
        data["gtensor_total"] = []
        data["delta_g"] = []
        ionic = -1
        dat = None
        for line in lines:
            first = line.strip()
            if first == "G tensor":
                ionic += 1
                for d in data.values():
                    d.append([])
            elif first in data:
                dat = first
            elif first.startswith("delta_g"):
                dat = "delta_g"
            else:
                splt = [postprocessor(s) for s in line.split()]
                splt = [s for s in splt if isinstance(s, float)]
                data[dat][ionic].append(list(map(float, splt[-3:])))
        self.data.update(data)
        return data["gtensor_total"][-1]

    def parse_chi_tensor(self, chi_filename=None):
        """Parse the magnetic susceptibility tensor."""
        if not chi_filename:
            if self.filenames["chi_tensor"]:
                chi_filename = self.filenames["chi_tensor"][0]
            else:
                return None

        with zopen(chi_filename, "rt") as f:
            lines = [line for line in f.read().split("\n") if line]

        data = {}
        data["chi_soft"] = []
        data["chi_local"] = []
        data["chi_total"] = []
        data["chi_total_ppm_cgs"] = []
        data["PV1"] = []
        data["PV2"] = []
        data["PV3"] = []
        data["ISO"] = []
        data["ANISO"] = []
        ionic = -1
        dat = None
        for line in lines:
            first = line.strip()
            if first == "Magnetic Susceptibility Tensor":
                ionic += 1
                for d in data.values():
                    d.append([])
            elif first in data:
                dat = first
            elif "SOFT" in first:
                dat = "chi_soft"
            elif "LOCAL" in first:
                dat = "chi_local"
            elif "Total" in first:
                dat = "chi_total_ppm_cgs" if "ppm" in first else "chi_total"
            elif first.startswith("PV1"):
                splt = [postprocessor(s) for s in line.split()]
                splt = [s for s in splt if isinstance(s, float)]
                data["PV1"][ionic] = splt[0]
                data["PV2"][ionic] = splt[1]
                data["PV3"][ionic] = splt[2]
            elif first.startswith("ISO"):
                splt = [postprocessor(s) for s in line.split()]
                splt = [s for s in splt if isinstance(s, float)]
                data["ISO"][ionic] = splt[0]
                data["ANISO"][ionic] = splt[1]
            else:
                splt = [postprocessor(s) for s in line.split()]
                splt = [s for s in splt if isinstance(s, float)]
                data[dat][ionic].append(list(map(float, splt)))
        self.data.update(data)
        return data["chi_total"][-1]

    def parse_nmr_shift(self):
        """Parse NMR calculation."""
        raise NotImplementedError("NMR Parsing not yet implemented")

    def parse_tddfpt(self):
        """Parse TDDFPT calculation."""
        raise NotImplementedError("TDDFPT excited states parsing not yet implemented")

    def parse_raman(self):
        """Parse raman calculation."""
        raise NotImplementedError("Raman parsing not yet implemented")

    @staticmethod
    def _gauss_smear(densities, energies, npts, width):
        if not width:
            return densities

        """Return a gaussian smeared DOS"""
        d = np.zeros(npts)
        e_s = np.linspace(min(energies), max(energies), npts)

        for e, _pd in zip(energies, densities):
            weight = np.exp(-(((e_s - e) / width) ** 2)) / (np.sqrt(np.pi) * width)
            d += _pd * weight

        return d

    def read_pattern(self, patterns, reverse=False, terminate_on_match=False, postprocess=str):
        r"""
        This function originally comes from pymatgen.io.vasp.outputs Outcar class.

        General pattern reading. Uses monty's regrep method. Takes the same
        arguments.

        Args:
            patterns (dict): A dict of patterns, e.g.,
                {"energy": r"energy\\(sigma->0\\)\\s+=\\s+([\\d\\-.]+)"}.
            reverse (bool): Read files in reverse. Defaults to false. Useful for
                large files, esp OUTCARs, especially when used with
                terminate_on_match.
            terminate_on_match (bool): Whether to terminate when there is at
                least one match in each key in pattern.
            postprocess (callable): A post processing function to convert all
                matches. Defaults to str, i.e., no change.

        Renders accessible:
            Any attribute in patterns. For example,
            {"energy": r"energy\\(sigma->0\\)\\s+=\\s+([\\d\\-.]+)"} will set the
            value of self.data["energy"] = [[-1234], [-3453], ...], to the
            results from regex and postprocess. Note that the returned values
            are lists of lists, because you can grep multiple items on one line.
        """
        matches = regrep(
            self.filename,
            patterns,
            reverse=reverse,
            terminate_on_match=terminate_on_match,
            postprocess=postprocess,
        )
        for k in patterns:
            self.data[k] = [i[0] for i in matches.get(k, [])]

    def read_table_pattern(
        self,
        header_pattern,
        row_pattern,
        footer_pattern,
        postprocess=str,
        attribute_name=None,
        last_one_only=True,
        strip=None,
    ):
        r"""
        This function originally comes from pymatgen.io.vasp.outputs Outcar class.

        Parse table-like data. A table composes of three parts: header,
        main body, footer. All the data matches "row pattern" in the main body
        will be returned.

        Args:
            header_pattern (str): The regular expression pattern matches the
                table header. This pattern should match all the text
                immediately before the main body of the table. For multiple
                sections table match the text until the section of
                interest. MULTILINE and DOTALL options are enforced, as a
                result, the "." meta-character will also match "\n" in this
                section.
            row_pattern (str): The regular expression matches a single line in
                the table. Capture interested field using regular expression
                groups.
            footer_pattern (str): The regular expression matches the end of the
                table. E.g. a long dash line.
            postprocess (callable): A post processing function to convert all
                matches. Defaults to str, i.e., no change.
            attribute_name (str): Name of this table. If present the parsed data
                will be attached to "data. e.g. self.data["efg"] = [...]
            last_one_only (bool): All the tables will be parsed, if this option
                is set to True, only the last table will be returned. The
                enclosing list will be removed. i.e. Only a single table will
                be returned. Default to be True.
            strip (list): Whether or not to strip contents out of the file before
                reading for a table pattern. This is mainly used by parse_scf_opt(),
                to strip HFX info out of the SCF loop start or DFT+U warnings out
                of the SCF loop iterations.

        Returns:
            List of tables. 1) A table is a list of rows. 2) A row if either a list of
            attribute values in case the the capturing group is defined without name in
            row_pattern, or a dict in case that named capturing groups are defined by
            row_pattern.
        """
        with zopen(self.filename, "rt") as f:
            if strip:
                lines = f.readlines()
                text = "".join(
                    [
                        lines[i]
                        for i in range(1, len(lines) - 1)
                        if all(
                            not lines[i].strip().startswith(c)
                            and not lines[i - 1].strip().startswith(c)
                            and not lines[i + 1].strip().startswith(c)
                            for c in strip
                        )
                    ]
                )
            else:
                text = f.read()

        table_pattern_text = header_pattern + r"\s*^(?P<table_body>(?:\s+" + row_pattern + r")+)\s+" + footer_pattern
        table_pattern = re.compile(table_pattern_text, re.MULTILINE | re.DOTALL)
        rp = re.compile(row_pattern)
        tables = []
        for mt in table_pattern.finditer(text):
            table_body_text = mt.group("table_body")
            table_contents = []
            for line in table_body_text.split("\n"):
                ml = rp.search(line)
                if ml is None:
                    continue
                d = ml.groupdict()
                if len(d) > 0:
                    processed_line = {k: postprocess(v) for k, v in d.items()}
                else:
                    processed_line = [postprocess(v) for v in ml.groups()]
                table_contents.append(processed_line)
            tables.append(table_contents)
        retained_data = tables[-1] if last_one_only else tables
        if attribute_name is not None:
            self.data[attribute_name] = retained_data
        return retained_data

    def as_dict(self):
        """Return dictionary representation of the output."""
        d = {"input": {}, "output": {}}
        d["total_time"] = self.timing["CP2K"]["total_time"]["maximum"]
        d["run_type"] = self.run_type
        d["input"]["global"] = self.data.get("global")
        d["input"]["dft"] = self.data.get("dft")
        d["input"]["scf"] = self.data.get("scf")
        d["input"]["qs"] = self.data.get("QS")
        d["input"]["run_type"] = self.run_type
        d["input"]["calculation_type"] = self.calculation_type
        d["input"]["structure"] = self.initial_structure.as_dict()
        d["input"]["atomic_kind_info"] = self.data.get("atomic_kind_info")
        d["input"]["cp2k_input"] = self.input.as_dict()
        d["ran_successfully"] = self.completed
        d["cp2k_version"] = self.cp2k_version
        d["output"]["structure"] = self.final_structure.as_dict()
        d["output"]["forces"] = self.data.get("forces", [None])[-1]
        d["output"]["stress"] = self.data.get("stress_tensor", [None])[-1]
        d["output"]["ionic_steps"] = [
            {k: v.as_dict() if isinstance(v, MSONable) else v for k, v in step.items()} for step in self.ionic_steps
        ]
        d["composition"] = self.composition.as_dict()
        d["output"]["energy"] = self.final_energy
        d["output"]["energy_per_atom"] = self.final_energy / self.composition.num_atoms
        d["output"]["bandgap"] = self.cbm - self.vbm if self.cbm and self.vbm else None
        d["output"]["cbm"] = self.cbm
        d["output"]["vbm"] = self.vbm
        d["output"]["efermi"] = self.efermi
        d["output"]["is_metal"] = self.is_metal
        return d


# TODO should store as pandas? Maybe it should be stored as a dict so it's python native
def parse_energy_file(energy_file):
    """Parses energy file for calculations with multiple ionic steps."""
    columns = [
        "step",
        "kinetic_energy",
        "temp",
        "potential_energy",
        "conserved_quantity",
        "used_time",
    ]
    df = pd.read_csv(energy_file, skiprows=1, names=columns, sep=r"\s+")
    df["kinetic_energy"] = df["kinetic_energy"] * Ha_to_eV
    df["potential_energy"] = df["potential_energy"] * Ha_to_eV
    df["conserved_quantity"] = df["conserved_quantity"] * Ha_to_eV
    df.astype(float)
    return {c: df[c].to_numpy() for c in columns}


# TODO The DOS file that cp2k outputs as of 2022.1 seems to have a lot of problems.
def parse_dos(dos_file=None):
    """Parse a dos file. This format is different from the pdos files."""
    data = np.loadtxt(dos_file)
    data[:, 0] *= Ha_to_eV
    energies = data[:, 0]
    for i, o in enumerate(data[:, 1]):
        if o == 0:
            break
        vbmtop = i
    efermi = energies[vbmtop] + 1e-6
    densities = {Spin.up: data[:, 1]}
    if data.shape[1] > 3:
        densities[Spin.down] = data[:, 3]
    return Dos(efermi=efermi, energies=energies, densities=densities)


def parse_pdos(dos_file=None, spin_channel=None, total=False):
    """
    Parse a single DOS file created by cp2k. Must contain one PDOS snapshot. i.e. you cannot
    use this cannot deal with multiple concatenated dos files.

    Args:
        dos_file (list): list of pdos_ALPHA file paths
        spin_channel (int): Which spin channel the file corresponds to. By default, CP2K will
            write the file with ALPHA or BETA in the filename (for spin up or down), but
            you can specify this here, in case you have a manual file name.
            spin_channel == 1 --> spin up, spin_channel == -1 --> spin down.
        total (bool): Whether to grab the total occupations, or the orbital decomposed ones.
        sigma (float): width for gaussian smearing, if desired

    Returns:
        Everything necessary to create a dos object, in dict format:
            (1) orbital decomposed DOS dict:
                i.e. pdoss = {specie: {orbital.s: {Spin.up: ... }, orbital.px: {Spin.up: ... } ...}}
            (2) energy levels of this dos file
            (3) fermi energy (in eV).
        DOS object is not created here

    """
    spin = Spin(spin_channel) if spin_channel else Spin.down if "BETA" in os.path.split(dos_file)[-1] else Spin.up

    with zopen(dos_file, "rt") as f:
        lines = f.readlines()
        kind = re.search(r"atomic kind\s(.*)\sat iter", lines[0]) or re.search(r"list\s(\d+)\s(.*)\sat iter", lines[0])
        kind = kind.groups()[0]

        header = re.split(r"\s{2,}", lines[1].replace("#", "").strip())[2:]
        dat = np.loadtxt(dos_file)

        def cp2k_to_pmg_labels(x):
            if x == "p":
                return "px"
            if x == "d":
                return "dxy"
            if x == "f":
                return "f_3"
            if x == "d-2":
                return "dxy"
            if x == "d-1":
                return "dyz"
            if x == "d0":
                return "dz2"
            if x == "d+1":
                return "dxz"
            if x == "d+2":
                return "dx2"
            if x == "f-3":
                return "f_3"
            if x == "f-2":
                return "f_2"
            if x == "f-1":
                return "f_1"
            if x == "f0":
                return "f0"
            if x == "f+1":
                return "f1"
            if x == "f+2":
                return "f2"
            if x == "f+3":
                return "f3"
            return x

        header = [cp2k_to_pmg_labels(h) for h in header]

        data = np.delete(dat, 0, 1)
        occupations = data[:, 1]
        data = np.delete(data, 1, 1)
        data[:, 0] *= Ha_to_eV
        energies = data[:, 0]
        for i, o in enumerate(occupations):
            if o == 0:
                break
            vbmtop = i

        # set Fermi level to be vbm plus tolerance for
        # PMG compatibility
        # *not* middle of the gap, which pdos might report
        efermi = energies[vbmtop] + 1e-6

        # for pymatgen's dos class. VASP creates an evenly spaced grid of energy states, which
        # leads to 0 density states in the band gap. CP2K does not do this. PMG's Dos class was
        # created with VASP in mind so the way it searches for vbm and cbm relies on grid points
        # in between VBM and CBM, so here we introduce trivial ones
        energies = np.insert(
            energies,
            vbmtop + 1,
            np.linspace(energies[vbmtop] + 1e-6, energies[vbmtop + 1] - 1e-6, 2),
        )
        data = np.insert(data, vbmtop + 1, np.zeros((2, data.shape[1])), axis=0)

        pdos = {
            kind: {
                getattr(Orbital, h): Dos(efermi=efermi, energies=energies, densities={spin: data[:, i + 1]})
                for i, h in enumerate(header)
            }
        }
        if total:
            tdos = Dos(efermi=efermi, energies=energies, densities={spin: np.sum(data[:, 1:], axis=1)})
            return pdos, tdos
        return pdos



"""
This package implements modules for input and output to and from Lobster. It
imports the key classes form both lobster.inputs and lobster_outputs to allow most
classes to be simply called as pymatgen.io.lobster.Lobsterin for example, to retain
backwards compatibility.
"""

from __future__ import annotations

from .inputs import Lobsterin
from .outputs import (
    Bandoverlaps,
    Charge,
    Cohpcar,
    Doscar,
    Fatband,
    Grosspop,
    Icohplist,
    Lobsterout,
    MadelungEnergies,
    SitePotential,
    Wavefunction,
)



# pylint: disable=C0414,W0718,C0301
# ruff: noqa: PLC0414
"""This package contains core modules and classes for representing structures and operations on them."""

from __future__ import annotations

import os
import warnings
from typing import Any

from ruamel.yaml import YAML

from pymatgen.core.composition import Composition as Composition
from pymatgen.core.lattice import Lattice as Lattice
from pymatgen.core.operations import SymmOp as SymmOp
from pymatgen.core.periodic_table import DummySpecie as DummySpecie
from pymatgen.core.periodic_table import DummySpecies as DummySpecies
from pymatgen.core.periodic_table import Element as Element
from pymatgen.core.periodic_table import Species as Species
from pymatgen.core.sites import PeriodicSite as PeriodicSite
from pymatgen.core.sites import Site as Site
from pymatgen.core.structure import IMolecule as IMolecule
from pymatgen.core.structure import IStructure as IStructure
from pymatgen.core.structure import Molecule as Molecule
from pymatgen.core.structure import SiteCollection as SiteCollection
from pymatgen.core.structure import Structure as Structure
from pymatgen.core.units import ArrayWithUnit as ArrayWithUnit
from pymatgen.core.units import FloatWithUnit as FloatWithUnit
from pymatgen.core.units import Unit as Unit

__author__ = "Pymatgen Development Team"
__email__ = "pymatgen@googlegroups.com"
__maintainer__ = "Shyue Ping Ong, Matthew Horton, Janosh Riebesell"
__maintainer_email__ = "shyuep@gmail.com"
__version__ = "2023.9.2"


SETTINGS_FILE = os.path.join(os.path.expanduser("~"), ".config", ".pmgrc.yaml")
OLD_SETTINGS_FILE = os.path.join(os.path.expanduser("~"), ".pmgrc.yaml")
MODULE_DIR = os.path.dirname(os.path.abspath(__file__))
PKG_DIR = os.path.dirname(MODULE_DIR)
ROOT = os.path.dirname(PKG_DIR)


def _load_pmg_settings() -> dict[str, Any]:
    settings: dict[str, Any] = {}

    # Load .pmgrc.yaml file
    yaml = YAML()
    for file_path in (SETTINGS_FILE, OLD_SETTINGS_FILE):
        try:
            with open(file_path, encoding="utf-8") as yml_file:
                settings = yaml.load(yml_file) or {}
            break
        except FileNotFoundError:
            continue
        except Exception as exc:
            # If there are any errors, default to using environment variables
            # if present.
            warnings.warn(f"Error loading {file_path}: {exc}.\nYou may need to reconfigure your yaml file.")

    # Override .pmgrc.yaml with env vars (if present)
    for key, val in os.environ.items():
        if key.startswith("PMG_"):
            settings[key] = val
        elif key in ("VASP_PSP_DIR", "MAPI_KEY", "DEFAULT_FUNCTIONAL"):
            settings[f"PMG_{key}"] = val

    return settings


SETTINGS = _load_pmg_settings()
locals().update(SETTINGS)



"""This module provides classes that define a chemical reaction."""

from __future__ import annotations

import logging
import re
from itertools import chain, combinations

import numpy as np
from monty.fractions import gcd_float
from monty.json import MontyDecoder, MSONable
from uncertainties import ufloat

from pymatgen.core.composition import Composition
from pymatgen.entries.computed_entries import ComputedEntry

__author__ = "Shyue Ping Ong, Anubhav Jain"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "2.0"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyuep@gmail.com"
__status__ = "Production"
__date__ = "Jul 11 2012"


logger = logging.getLogger(__name__)


class BalancedReaction(MSONable):
    """An object representing a complete chemical reaction."""

    # Tolerance for determining if a particular component fraction is > 0.
    TOLERANCE = 1e-6

    def __init__(self, reactants_coeffs, products_coeffs):
        """
        Reactants and products to be specified as dict of {Composition: coeff}.

        Args:
            reactants_coeffs (dict[Composition, float]): Reactants as dict of {Composition: amt}.
            products_coeffs (dict[Composition, float]): Products as dict of {Composition: amt}.
        """
        # sum reactants and products
        all_reactants = sum((k * v for k, v in reactants_coeffs.items()), Composition({}))
        all_products = sum((k * v for k, v in products_coeffs.items()), Composition({}))

        if not all_reactants.almost_equals(all_products, rtol=0, atol=self.TOLERANCE):
            raise ReactionError("Reaction is unbalanced!")

        self._els = all_reactants.elements

        self.reactants_coeffs = reactants_coeffs
        self.products_coeffs = products_coeffs

        # calculate net reaction coefficients
        self._coeffs = []
        self._els = []
        self._all_comp = []
        for key in {*reactants_coeffs, *products_coeffs}:
            coeff = products_coeffs.get(key, 0) - reactants_coeffs.get(key, 0)

            if abs(coeff) > self.TOLERANCE:
                self._all_comp.append(key)
                self._coeffs.append(coeff)

    def calculate_energy(self, energies):
        """
        Calculates the energy of the reaction.

        Args:
            energies ({Composition: float}): Energy for each composition.
                E.g ., {comp1: energy1, comp2: energy2}.

        Returns:
            reaction energy as a float.
        """
        return sum(amt * energies[c] for amt, c in zip(self._coeffs, self._all_comp))

    def normalize_to(self, comp, factor=1):
        """
        Normalizes the reaction to one of the compositions.
        By default, normalizes such that the composition given has a
        coefficient of 1. Another factor can be specified.

        Args:
            comp (Composition): Composition to normalize to
            factor (float): Factor to normalize to. Defaults to 1.
        """
        scale_factor = abs(1 / self._coeffs[self._all_comp.index(comp)] * factor)
        self._coeffs = [c * scale_factor for c in self._coeffs]

    def normalize_to_element(self, element, factor=1):
        """
        Normalizes the reaction to one of the elements.
        By default, normalizes such that the amount of the element is 1.
        Another factor can be specified.

        Args:
            element (Element/Species): Element to normalize to.
            factor (float): Factor to normalize to. Defaults to 1.
        """
        all_comp = self._all_comp
        coeffs = self._coeffs
        current_el_amount = sum(all_comp[i][element] * abs(coeffs[i]) for i in range(len(all_comp))) / 2
        scale_factor = factor / current_el_amount
        self._coeffs = [c * scale_factor for c in coeffs]

    def get_el_amount(self, element):
        """
        Returns the amount of the element in the reaction.

        Args:
            element (Element/Species): Element in the reaction

        Returns:
            Amount of that element in the reaction.
        """
        return sum(self._all_comp[i][element] * abs(self._coeffs[i]) for i in range(len(self._all_comp))) / 2

    @property
    def elements(self):
        """List of elements in the reaction."""
        return self._els[:]

    @property
    def coeffs(self):
        """Final coefficients of the calculated reaction."""
        return self._coeffs[:]

    @property
    def all_comp(self):
        """List of all compositions in the reaction."""
        return self._all_comp

    @property
    def reactants(self):
        """List of reactants."""
        return [self._all_comp[i] for i in range(len(self._all_comp)) if self._coeffs[i] < 0]

    @property
    def products(self):
        """List of products."""
        return [self._all_comp[i] for i in range(len(self._all_comp)) if self._coeffs[i] > 0]

    def get_coeff(self, comp):
        """Returns coefficient for a particular composition."""
        return self._coeffs[self._all_comp.index(comp)]

    def normalized_repr_and_factor(self):
        """
        Normalized representation for a reaction
        For example, ``4 Li + 2 O -> 2Li2O`` becomes ``2 Li + O -> Li2O``.
        """
        return self._str_from_comp(self._coeffs, self._all_comp, True)

    @property
    def normalized_repr(self):
        """
        A normalized representation of the reaction. All factors are converted
        to lowest common factors.
        """
        return self.normalized_repr_and_factor()[0]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        for comp in self._all_comp:
            coeff2 = other.get_coeff(comp) if comp in other._all_comp else 0
            if abs(self.get_coeff(comp) - coeff2) > self.TOLERANCE:
                return False
        return True

    def __hash__(self) -> int:
        return 7

    @classmethod
    def _str_from_formulas(cls, coeffs, formulas):
        reactant_str = []
        product_str = []
        for amt, formula in zip(coeffs, formulas):
            if abs(amt + 1) < cls.TOLERANCE:
                reactant_str.append(formula)
            elif abs(amt - 1) < cls.TOLERANCE:
                product_str.append(formula)
            elif amt < -cls.TOLERANCE:
                reactant_str.append(f"{-amt:.4g} {formula}")
            elif amt > cls.TOLERANCE:
                product_str.append(f"{amt:.4g} {formula}")

        return f"{' + '.join(reactant_str)} -> {' + '.join(product_str)}"

    @classmethod
    def _str_from_comp(cls, coeffs, compositions, reduce=False):
        r_coeffs = np.zeros(len(coeffs))
        r_formulas = []
        for i, (amt, comp) in enumerate(zip(coeffs, compositions)):
            formula, factor = comp.get_reduced_formula_and_factor()
            r_coeffs[i] = amt * factor
            r_formulas.append(formula)
        if reduce:
            factor = 1 / gcd_float(np.abs(r_coeffs))
            r_coeffs *= factor
        else:
            factor = 1
        return cls._str_from_formulas(r_coeffs, r_formulas), factor

    def __str__(self):
        return self._str_from_comp(self._coeffs, self._all_comp)[0]

    __repr__ = __str__

    def as_entry(self, energies):
        """
        Returns a ComputedEntry representation of the reaction.
        """
        relevant_comp = [comp * abs(coeff) for coeff, comp in zip(self._coeffs, self._all_comp)]
        comp = sum(relevant_comp, Composition())
        entry = ComputedEntry(0.5 * comp, self.calculate_energy(energies))
        entry.name = str(self)
        return entry

    def as_dict(self):
        """
        Returns:
            A dictionary representation of BalancedReaction.
        """
        return {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "reactants": {str(comp): coeff for comp, coeff in self.reactants_coeffs.items()},
            "products": {str(comp): coeff for comp, coeff in self.products_coeffs.items()},
        }

    @classmethod
    def from_dict(cls, d):
        """
        Args:
            d (dict): from as_dict().

        Returns:
            A BalancedReaction object.
        """
        reactants = {Composition(comp): coeff for comp, coeff in d["reactants"].items()}
        products = {Composition(comp): coeff for comp, coeff in d["products"].items()}
        return cls(reactants, products)

    @classmethod
    @np.deprecate(message="Use from_str instead")
    def from_string(cls, *args, **kwargs):
        return cls.from_str(*args, **kwargs)

    @staticmethod
    def from_str(rxn_string):
        """
        Generates a balanced reaction from a string. The reaction must
        already be balanced.

        Args:
            rxn_string:
                The reaction string. For example, "4 Li + O2-> 2Li2O"

        Returns:
            BalancedReaction
        """
        rct_str, prod_str = rxn_string.split("->")

        def get_comp_amt(comp_str):
            return {
                Composition(m.group(2)): float(m.group(1) or 1)
                for m in re.finditer(r"([\d\.]*(?:[eE]-?[\d\.]+)?)\s*([A-Z][\w\.\(\)]*)", comp_str)
            }

        return BalancedReaction(get_comp_amt(rct_str), get_comp_amt(prod_str))


class Reaction(BalancedReaction):
    """
    A more flexible class representing a Reaction. The reaction amounts will
    be automatically balanced. Reactants and products can swap sides so that
    all coefficients are positive, however this class will find the solution
    with the minimum number of swaps and coefficients of 0. Normalizes so that
    the *FIRST* product (or products, if underdetermined) has a coefficient of one.
    """

    def __init__(self, reactants, products):
        """
        Reactants and products to be specified as list of
        pymatgen.core.structure.Composition. e.g., [comp1, comp2].

        Args:
            reactants ([Composition]): List of reactants.
            products ([Composition]): List of products.
        """
        self._input_reactants = reactants
        self._input_products = products
        self._all_comp = reactants + products
        self._num_comp = len(self.all_comp)

        all_elems = sorted({elem for c in self._all_comp for elem in c.elements})
        self._num_elems = len(all_elems)

        comp_matrix = np.array([[c[el] for el in all_elems] for c in self._all_comp]).T

        rank = np.linalg.matrix_rank(comp_matrix)
        diff = self._num_comp - rank
        num_constraints = diff if diff >= 2 else 1

        self._lowest_num_errors = np.inf  # an error = a component changing sides or disappearing

        self._coeffs = self._balance_coeffs(comp_matrix, num_constraints)
        self._els = all_elems

    def _balance_coeffs(self, comp_matrix, max_num_constraints):
        first_product_idx = len(self._input_reactants)

        # start with simplest product constraints, work towards most complex reactant constraints
        product_constraints = chain.from_iterable(
            [
                combinations(range(first_product_idx, self._num_comp), n_constr)
                for n_constr in range(max_num_constraints, 0, -1)
            ]
        )
        reactant_constraints = chain.from_iterable(
            [combinations(range(first_product_idx), n_constr) for n_constr in range(max_num_constraints, 0, -1)]
        )
        best_soln = None
        balanced = False

        for constraints in chain(product_constraints, reactant_constraints):
            n_constr = len(constraints)

            comp_and_constraints = np.append(comp_matrix, np.zeros((n_constr, self._num_comp)), axis=0)
            b = np.zeros((self._num_elems + n_constr, 1))
            b[-n_constr:] = 1 if min(constraints) >= first_product_idx else -1

            for num, idx in enumerate(constraints):
                comp_and_constraints[self._num_elems + num, idx] = 1
                # arbitrarily fix coeff to 1

            coeffs = np.matmul(np.linalg.pinv(comp_and_constraints), b)

            if np.allclose(np.matmul(comp_matrix, coeffs), np.zeros((self._num_elems, 1))):
                balanced = True
                expected_signs = np.array([-1] * len(self._input_reactants) + [+1] * len(self._input_products))
                num_errors = np.sum(np.multiply(expected_signs, coeffs.T) < self.TOLERANCE)

                if num_errors == 0:
                    self._lowest_num_errors = 0
                    return np.squeeze(coeffs)
                if num_errors < self._lowest_num_errors:
                    self._lowest_num_errors = num_errors
                    best_soln = coeffs

        if not balanced:
            raise ReactionError("Reaction cannot be balanced.")

        return np.squeeze(best_soln)

    def copy(self):
        """Returns a copy of the Reaction object."""
        return Reaction(self.reactants, self.products)

    def as_dict(self):
        """
        Returns:
            A dictionary representation of Reaction.
        """
        return {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "reactants": [comp.as_dict() for comp in self._input_reactants],
            "products": [comp.as_dict() for comp in self._input_products],
        }

    @classmethod
    def from_dict(cls, d):
        """
        Args:
            d (dict): from as_dict().

        Returns:
            A Reaction object.
        """
        reactants = [Composition(sym_amt) for sym_amt in d["reactants"]]
        products = [Composition(sym_amt) for sym_amt in d["products"]]
        return cls(reactants, products)


class ReactionError(Exception):
    """
    Exception class for Reactions. Allows more information in exception
    messages to cover situations not covered by standard exception classes.
    """

    def __init__(self, msg):
        """
        Create a ReactionError.

        Args:
            msg (str): More information about the ReactionError.
        """
        self.msg = msg

    def __str__(self):
        return self.msg


class ComputedReaction(Reaction):
    """
    Convenience class to generate a reaction from ComputedEntry objects, with
    some additional attributes, such as a reaction energy based on computed
    energies.
    """

    def __init__(self, reactant_entries, product_entries):
        """
        Args:
            reactant_entries ([ComputedEntry]): List of reactant_entries.
            product_entries ([ComputedEntry]): List of product_entries.
        """
        self._reactant_entries = reactant_entries
        self._product_entries = product_entries
        self._all_entries = reactant_entries + product_entries
        reactant_comp = [e.composition.reduced_composition for e in reactant_entries]

        product_comp = [e.composition.reduced_composition for e in product_entries]

        super().__init__(list(reactant_comp), list(product_comp))

    @property
    def all_entries(self):
        """
        Equivalent of all_comp but returns entries, in the same order as the
        coefficients.
        """
        entries = []
        for c in self._all_comp:
            for e in self._all_entries:
                if e.composition.reduced_formula == c.reduced_formula:
                    entries.append(e)
                    break
        return entries

    @property
    def calculated_reaction_energy(self):
        """
        Returns (float):
            The calculated reaction energy.
        """
        calc_energies = {}

        for entry in self._reactant_entries + self._product_entries:
            comp, factor = entry.composition.get_reduced_composition_and_factor()
            calc_energies[comp] = min(calc_energies.get(comp, float("inf")), entry.energy / factor)

        return self.calculate_energy(calc_energies)

    @property
    def calculated_reaction_energy_uncertainty(self):
        """
        Calculates the uncertainty in the reaction energy based on the uncertainty in the
        energies of the products and reactants.
        """
        calc_energies = {}

        for entry in self._reactant_entries + self._product_entries:
            comp, factor = entry.composition.get_reduced_composition_and_factor()
            energy_ufloat = ufloat(entry.energy, entry.correction_uncertainty)
            calc_energies[comp] = min(calc_energies.get(comp, float("inf")), energy_ufloat / factor)

        return self.calculate_energy(calc_energies).std_dev

    def as_dict(self):
        """
        Returns:
            A dictionary representation of ComputedReaction.
        """
        return {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "reactants": [e.as_dict() for e in self._reactant_entries],
            "products": [e.as_dict() for e in self._product_entries],
        }

    @classmethod
    def from_dict(cls, d):
        """
        Args:
            d (dict): from as_dict().

        Returns:
            A ComputedReaction object.
        """
        dec = MontyDecoder()
        reactants = [dec.process_decoded(e) for e in d["reactants"]]
        products = [dec.process_decoded(e) for e in d["products"]]
        return cls(reactants, products)



"""Defines an abstract base class contract for Transformation object."""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from monty.json import MSONable

if TYPE_CHECKING:
    from pymatgen.core import Structure

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyuep@gmail.com"
__date__ = "Sep 23, 2011"


class AbstractTransformation(MSONable, metaclass=abc.ABCMeta):
    """Abstract transformation class."""

    @abc.abstractmethod
    def apply_transformation(self, structure: Structure):
        """Applies the transformation to a structure. Depending on whether a
        transformation is one-to-many, there may be an option to return a
        ranked list of structures.

        Args:
            structure:
                input structure
            return_ranked_list (bool | int, optional): If return_ranked_list is int, that number of structures

                is returned. If False, only the single lowest energy structure is returned. Defaults to False.

        Returns:
            depending on returned_ranked list, either a transformed structure
            or
            a list of dictionaries, where each dictionary is of the form
            {'structure' = .... , 'other_arguments'}
            the key 'transformation' is reserved for the transformation that
            was actually applied to the structure.
            This transformation is parsed by the alchemy classes for generating
            a more specific transformation history. Any other information will
            be stored in the transformation_parameters dictionary in the
            transmuted structure class.
        """
        return

    @property
    @abc.abstractmethod
    def inverse(self) -> AbstractTransformation | None:
        """Returns the inverse transformation if available.
        Otherwise, should return None.
        """

    @property
    @abc.abstractmethod
    def is_one_to_many(self) -> bool:
        """Determines if a Transformation is a one-to-many transformation. If a
        Transformation is a one-to-many transformation, the
        apply_transformation method should have a keyword arg
        "return_ranked_list" which allows for the transformed structures to be
        returned as a ranked list.
        """
        return False

    @property
    def use_multiprocessing(self) -> bool:
        """Indicates whether the transformation can be applied by a
        subprocessing pool. This should be overridden to return True for
        transformations that the transmuter can parallelize.
        """
        return False
